# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Any, Union

import numpy as np
import pyproj
import shapely
import shapely.ops
import xarray as xr

from xcube.constants import CRS_CRS84, LOG
from xcube.core.geom import (
    get_dataset_geometry,
    mask_dataset_by_geometry,
    normalize_geometry,
)
from xcube.core.varexpr import VarExprContext, split_var_assignment
from xcube.server.api import ApiError
from xcube.util.perf import measure_time_cm

from .context import StatisticsContext

NAN_RESULT = {"count": 0}
NAN_RESULT_COMPACT = {}
DEFAULT_BIN_COUNT = 100


def compute_statistics(
    ctx: StatisticsContext,
    ds_id: str,
    var_name: str,
    geometry: Union[dict[str, Any], tuple[float, float]],
    time_label: str,
    trace_perf: bool = False,
):
    measure_time = measure_time_cm(logger=LOG, disabled=not trace_perf)
    with measure_time("Computing statistics"):
        return _compute_statistics(
            ctx, ds_id, var_name, time_label, geometry, DEFAULT_BIN_COUNT
        )


def _compute_statistics(
    ctx: StatisticsContext,
    ds_id: str,
    var_name_or_assign: str,
    time_label: str,
    geometry: Union[dict[str, Any], tuple[float, float]],
    bin_count: int,
):
    ml_dataset = ctx.datasets_ctx.get_ml_dataset(ds_id)
    dataset = ml_dataset.get_dataset(0)
    grid_mapping = ml_dataset.grid_mapping

    dataset_contains_time = "time" in dataset

    if dataset_contains_time:
        if time_label is not None:
            try:
                time = np.array(time_label, dtype=dataset.time.dtype)
                dataset = dataset.sel(time=time, method="nearest")
            except (TypeError, ValueError) as e:
                raise ApiError.BadRequest("Invalid query parameter 'time'") from e
        else:
            raise ApiError.BadRequest("Missing query parameter 'time'")
    elif time_label is not None:
        raise ApiError.BadRequest(
            "Query parameter 'time' must not be given"
            " since dataset does not contain a 'time' dimension"
        )

    if isinstance(geometry, tuple):
        compact_mode = True
        geometry = shapely.geometry.Point(geometry)
    else:
        compact_mode = False
        try:
            geometry = normalize_geometry(geometry)
        except (TypeError, ValueError, AttributeError) as e:
            raise ApiError.BadRequest(
                f"Invalid GeoJSON geometry encountered: {{e}}"
            ) from e

    if geometry is not None and not grid_mapping.crs.is_geographic:
        project = pyproj.Transformer.from_crs(
            CRS_CRS84, grid_mapping.crs, always_xy=True
        ).transform
        geometry = shapely.ops.transform(project, geometry)

    nan_result = NAN_RESULT_COMPACT if compact_mode else NAN_RESULT

    x_name, y_name = grid_mapping.xy_dim_names
    if isinstance(geometry, shapely.geometry.Point):
        bounds = get_dataset_geometry(dataset)
        if not bounds.contains(geometry):
            return nan_result
        indexers = {x_name: geometry.x, y_name: geometry.y}
        variable = _get_dataset_variable(var_name_or_assign, dataset)
        value = variable.sel(**indexers, method="Nearest").values
        if np.isnan(value):
            return nan_result
        if compact_mode:
            return {"value": float(value)}
        else:
            return {
                "count": 1,
                "minimum": float(value),
                "maximum": float(value),
                "mean": float(value),
                "deviation": 0.0,
            }

    dataset = mask_dataset_by_geometry(dataset, geometry)
    if dataset is None:
        return nan_result

    variable = _get_dataset_variable(var_name_or_assign, dataset)

    count = int(np.count_nonzero(~np.isnan(variable)))
    if count == 0:
        return nan_result

    # note, casting to float forces intended computation
    minimum = float(variable.min())
    maximum = float(variable.max())
    h_values, h_edges = np.histogram(
        variable, bin_count, range=(minimum, maximum), density=True
    )

    return {
        "count": count,
        "minimum": minimum,
        "maximum": maximum,
        "mean": float(variable.mean()),
        "deviation": float(variable.std()),
        "histogram": {
            "values": [float(v) for v in h_values],
            "edges": [float(v) for v in h_edges],
        },
    }


def _get_dataset_variable(var_name_or_assign: str, dataset: xr.Dataset) -> xr.DataArray:
    var_name, var_expr = split_var_assignment(var_name_or_assign)
    if var_expr:
        variable = VarExprContext(dataset).evaluate(var_expr)
        variable.name = var_name
    else:
        var_name = var_name_or_assign
        variable = dataset[var_name]

    return variable



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import collections.abc
import numbers
from collections.abc import Iterable, Mapping, Sequence
from typing import Any, Optional, Tuple, Union

import pyproj

from xcube.util.assertions import assert_given, assert_instance, assert_true
from xcube.util.jsonschema import (
    JsonArraySchema,
    JsonBooleanSchema,
    JsonDateSchema,
    JsonIntegerSchema,
    JsonNumberSchema,
    JsonObject,
    JsonObjectSchema,
    JsonStringSchema,
)


class InputConfig(JsonObject):
    def __init__(
        self,
        store_id: str = None,
        opener_id: str = None,
        data_id: str = None,
        store_params: Mapping[str, Any] = None,
        open_params: Mapping[str, Any] = None,
    ):
        assert_true(
            store_id or opener_id, "One of store_id and opener_id must be given"
        )
        assert_given(data_id, "data_id")
        self.store_id = store_id
        self.opener_id = opener_id
        self.data_id = data_id
        self.store_params = store_params
        self.open_params = open_params

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                store_id=JsonStringSchema(min_length=1),
                opener_id=JsonStringSchema(min_length=1),
                data_id=JsonStringSchema(min_length=1),
                store_params=JsonObjectSchema(
                    additional_properties=True, nullable=True
                ),
                open_params=JsonObjectSchema(additional_properties=True, nullable=True),
            ),
            additional_properties=False,
            required=["data_id"],
            factory=cls,
        )


class CallbackConfig(JsonObject):
    def __init__(self, api_uri: str = None, access_token: str = None):
        assert_true(
            api_uri and access_token, "Both, api_uri and access_token must be given"
        )
        self.api_uri = api_uri
        self.access_token = access_token

    @classmethod
    def get_schema(cls):
        return JsonObjectSchema(
            properties=dict(
                api_uri=JsonStringSchema(min_length=1),
                access_token=JsonStringSchema(min_length=1),
            ),
            additional_properties=False,
            required=["api_uri", "access_token"],
            factory=cls,
        )


class OutputConfig(JsonObject):
    def __init__(
        self,
        store_id: str = None,
        writer_id: str = None,
        data_id: str = None,
        store_params: Mapping[str, Any] = None,
        write_params: Mapping[str, Any] = None,
        replace: bool = None,
    ):
        assert_true(
            store_id or writer_id, "One of store_id and writer_id must be given"
        )
        self.store_id = store_id
        self.writer_id = writer_id
        self.data_id = data_id
        self.store_params = store_params
        self.write_params = write_params
        self.replace = replace

    @classmethod
    def get_schema(cls):
        return JsonObjectSchema(
            properties=dict(
                store_id=JsonStringSchema(min_length=1),
                writer_id=JsonStringSchema(min_length=1),
                data_id=JsonStringSchema(default=None),
                store_params=JsonObjectSchema(
                    additional_properties=True, nullable=True
                ),
                write_params=JsonObjectSchema(
                    additional_properties=True, nullable=True
                ),
                replace=JsonBooleanSchema(default=False),
            ),
            additional_properties=False,
            required=[],
            factory=cls,
        )


# Need to be aligned with params in resample_cube(cube, **params)
class CubeConfig(JsonObject):
    def __init__(
        self,
        variable_names: Sequence[str] = None,
        crs: str = None,
        bbox: tuple[float, float, float, float] = None,
        spatial_res: Union[float, tuple[float]] = None,
        tile_size: Union[int, tuple[int, int]] = None,
        time_range: tuple[str, Optional[str]] = None,
        time_period: str = None,
        chunks: Mapping[str, Optional[int]] = None,
        metadata: Mapping[str, Any] = None,
        variable_metadata: Mapping[str, Mapping[str, Any]] = None,
    ):
        self.variable_names = None
        if variable_names is not None:
            assert_true(len(variable_names) > 0, "variable_names is invalid")
            self.variable_names = tuple(map(str, variable_names))

        self.crs = None
        if crs is not None:
            assert_instance(crs, str, "crs")
            try:
                pyproj.crs.CRS.from_string(crs)
            except pyproj.exceptions.CRSError:
                raise ValueError("crs is invalid")
            self.crs = crs

        self.bbox = None
        if bbox is not None:
            assert_true(len(bbox) == 4, "bbox is invalid")
            self.bbox = tuple(map(float, bbox))

        self.spatial_res = None
        if spatial_res is not None:
            assert_instance(spatial_res, numbers.Number, "spatial_res")
            self.spatial_res = float(spatial_res)

        self.tile_size = None
        if tile_size is not None:
            if isinstance(tile_size, int):
                tile_width, tile_height = tile_size, tile_size
            else:
                try:
                    tile_width, tile_height = tile_size
                except (ValueError, TypeError):
                    raise ValueError(
                        "tile_size must be an integer or a pair of integers"
                    )
                assert_instance(tile_width, numbers.Number, "tile_width of tile_size")
                assert_instance(tile_height, numbers.Number, "tile_height of tile_size")
            self.tile_size = tile_width, tile_height

        self.time_range = None
        if time_range is not None:
            assert_true(len(time_range) == 2, "time_range is invalid")
            self.time_range = tuple(time_range)

        self.time_period = None
        if time_period is not None:
            assert_instance(time_period, str, "time_period")
            self.time_period = time_period

        self.chunks = None
        if chunks is not None:
            assert_instance(chunks, collections.abc.Mapping, "chunks")
            for chunk_dim, chunk_size in chunks.items():
                assert_instance(chunk_dim, str, "chunk dimension")
                assert_instance(chunk_size, (int, type(None)), "chunk size")
            self.chunks = dict(chunks)

        self.metadata = None
        if metadata is not None:
            assert_instance(metadata, collections.abc.Mapping, "metadata")
            self.metadata = dict(metadata)

        self.variable_metadata = None
        if variable_metadata is not None:
            assert_instance(
                variable_metadata, collections.abc.Mapping, "variable_metadata"
            )
            for var_name, metadata in variable_metadata.items():
                assert_instance(var_name, str, "chunk dimension")
                assert_instance(metadata, collections.abc.Mapping, "variable metadata")
            self.variable_metadata = dict(variable_metadata)

    def drop_props(self, name: Union[str, Iterable[str]]):
        """Drop one or more named properties from this configuration.

        Args:
            name: the name of the property to be dropped.
            names: more names of properties to be dropped.

        Returns:
            a new cube configuration.
        """
        if not name:
            return self
        name_set = {name} if isinstance(name, str) else set(name)
        for k in name_set:
            assert_true(
                hasattr(self, k), message=f"{k} is not a property of {CubeConfig!r}"
            )
        return self.from_dict(
            {k: v for k, v in self.to_dict().items() if k not in name_set}
        )

    @classmethod
    def get_schema(cls):
        return JsonObjectSchema(
            properties=dict(
                variable_names=JsonArraySchema(
                    nullable=True, items=JsonStringSchema(min_length=1), min_items=0
                ),
                crs=JsonStringSchema(nullable=True, min_length=1),
                bbox=JsonArraySchema(
                    nullable=True,
                    items=[
                        JsonNumberSchema(),
                        JsonNumberSchema(),
                        JsonNumberSchema(),
                        JsonNumberSchema(),
                    ],
                ),
                spatial_res=JsonNumberSchema(nullable=True, exclusive_minimum=0.0),
                tile_size=JsonArraySchema(
                    nullable=True,
                    items=[
                        JsonIntegerSchema(minimum=1, maximum=2500),
                        JsonIntegerSchema(minimum=1, maximum=2500),
                    ],
                ),
                time_range=JsonDateSchema.new_range(nullable=True),
                time_period=JsonStringSchema(
                    nullable=True, pattern=r"^([1-9][0-9]*)?[DWMY]$"
                ),
                chunks=JsonObjectSchema(
                    nullable=True,
                    additional_properties=JsonIntegerSchema(nullable=True, minimum=1),
                ),
                metadata=JsonObjectSchema(nullable=True, additional_properties=True),
                variable_metadata=JsonObjectSchema(
                    nullable=True,
                    additional_properties=JsonObjectSchema(additional_properties=True),
                ),
            ),
            additional_properties=False,
            factory=cls,
        )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from collections.abc import Hashable, Mapping, Sequence
from typing import Dict, Optional, Tuple, Union

import numpy as np
import xarray as xr

from xcube.core.gridmapping import GridMapping


class CubeSchema:
    """A schema that can be used to create new xcube datasets.
    The given *shape*, *dims*, and *chunks*, *coords* apply to all data variables.

    Args:
        shape: A tuple of dimension sizes.
        coords: A dictionary of coordinate variables. Must have values
            for all *dims*.
        dims: A sequence of dimension names. Defaults to ``('time',
            'lat', 'lon')``.
        chunks: A tuple of chunk sizes in each dimension.
    """

    def __init__(
        self,
        shape: Sequence[int],
        coords: Mapping[str, xr.DataArray],
        x_name: str = "lon",
        y_name: str = "lat",
        time_name: str = "time",
        dims: Sequence[str] = None,
        chunks: Sequence[int] = None,
    ):
        if not shape:
            raise ValueError("shape must be a sequence of integer sizes")
        if not coords:
            raise ValueError(
                "coords must be a mapping from dimension names to label arrays"
            )
        if not x_name:
            raise ValueError("x_name must be given")
        if not y_name:
            raise ValueError("y_name must be given")
        if not time_name:
            raise ValueError("time_name must be given")

        ndim = len(shape)
        if ndim < 3:
            raise ValueError("shape must have at least three dimensions")
        dims = tuple(dims) or (time_name, y_name, x_name)
        if dims and len(dims) != ndim:
            raise ValueError("dims must have same length as shape")
        if x_name not in coords or y_name not in coords or time_name not in coords:
            raise ValueError(
                f"missing variables {x_name!r}, {y_name!r}, {time_name!r} in coords"
            )
        x_var, y_var, time_var = (
            coords.get(x_name),
            coords.get(y_name),
            coords.get(time_name),
        )
        if x_var.ndim != 1 or y_var.ndim != 1 or time_var.ndim != 1:
            raise ValueError(
                f"variables {x_name!r}, {y_name!r}, {time_name!r} in coords must be 1-D"
            )
        x_dim, y_dim, time_dim = x_var.dims[0], y_var.dims[0], time_var.dims[0]
        if dims[0] != time_dim:
            raise ValueError(f"the first dimension in dims must be {time_dim!r}")
        if dims[-2:] != (y_dim, x_dim):
            raise ValueError(
                f"the last two dimensions in dims must be {y_dim!r} and {x_dim!r}"
            )
        if chunks and len(chunks) != ndim:
            raise ValueError("chunks must have same length as shape")
        for i in range(ndim):
            dim_name = dims[i]
            dim_size = shape[i]
            if dim_name not in coords:
                raise ValueError(f"missing dimension {dim_name!r} in coords")
            dim_labels = coords[dim_name]
            if len(dim_labels.shape) != 1:
                raise ValueError(
                    f"labels of {dim_name!r} in coords must be one-dimensional"
                )
            if len(dim_labels) != dim_size:
                raise ValueError(
                    f"number of labels of {dim_name!r} in coords does not match shape"
                )

        self._shape = tuple(shape)
        self._x_name = x_name
        self._y_name = y_name
        self._time_name = time_name
        self._dims = dims
        self._chunks = tuple(chunks) if chunks else None
        self._coords = dict(coords)

    @property
    def ndim(self) -> int:
        """Number of dimensions."""
        return len(self._dims)

    @property
    def dims(self) -> tuple[str, ...]:
        """Tuple of dimension names."""
        return self._dims

    @property
    def x_name(self) -> str:
        """Name of the spatial x coordinate variable."""
        return self._x_name

    @property
    def y_name(self) -> str:
        """Name of the spatial y coordinate variable."""
        return self._y_name

    @property
    def time_name(self) -> str:
        """Name of the time coordinate variable."""
        return self._time_name

    @property
    def x_var(self) -> xr.DataArray:
        """Spatial x coordinate variable."""
        return self._coords[self._x_name]

    @property
    def y_var(self) -> xr.DataArray:
        """Spatial y coordinate variable."""
        return self._coords[self._y_name]

    @property
    def time_var(self) -> xr.DataArray:
        """Time coordinate variable."""
        return self._coords[self._time_name]

    @property
    def x_dim(self) -> str:
        """Name of the spatial x dimension."""
        return self._dims[-1]

    @property
    def y_dim(self) -> str:
        """Name of the spatial y dimension."""
        return self._dims[-2]

    @property
    def time_dim(self) -> str:
        """Name of the time dimension."""
        return self._dims[0]

    @property
    def x_size(self) -> int:
        """Size of the spatial x dimension."""
        return self._shape[-1]

    @property
    def y_size(self) -> int:
        """Size of the spatial y dimension."""
        return self._shape[-2]

    @property
    def time_size(self) -> int:
        """Size of the time dimension."""
        return self._shape[0]

    @property
    def shape(self) -> tuple[int, ...]:
        """Tuple of dimension sizes."""
        return self._shape

    @property
    def chunks(self) -> Optional[tuple[int]]:
        """Tuple of dimension chunk sizes."""
        return self._chunks

    @property
    def coords(self) -> dict[str, xr.DataArray]:
        """Dictionary of coordinate variables."""
        return self._coords

    @classmethod
    def new(cls, cube: xr.Dataset) -> "CubeSchema":
        """Create a cube schema from given *cube*."""
        return get_cube_schema(cube)

    def _repr_html_(self):
        """Return a HTML representation for Jupyter Notebooks."""
        return (
            f"<table>"
            f"<tr><td>Shape:</td><td>{self.shape}</td></tr>"
            f"<tr><td>Chunk sizes:</td><td>{self.chunks}</td></tr>"
            f"<tr><td>Dimensions:</td><td>{self.dims}</td></tr>"
            f"</table>"
        )


# TODO (forman): code duplication with xcube.core.verify._check_data_variables(), line 76
def get_cube_schema(cube: xr.Dataset) -> CubeSchema:
    """Derive cube schema from given *cube*.

    Args:
        cube: The data cube.

    Returns:
        The cube schema.
    """

    xy_var_names = get_dataset_xy_var_names(
        cube, must_exist=True, dataset_arg_name="cube"
    )
    time_var_name = get_dataset_time_var_name(
        cube, must_exist=True, dataset_arg_name="cube"
    )

    first_dims = None
    first_shape = None
    first_chunks = None
    first_coords = None

    for var_name, var in cube.data_vars.items():
        dims = var.dims
        if first_dims is None:
            first_dims = dims
        elif first_dims != dims:
            raise ValueError(
                f"all variables must have same dimensions, but variable {var_name!r} "
                f"has dimensions {dims!r}"
            )

        shape = var.shape
        if first_shape is None:
            first_shape = shape
        elif first_shape != shape:
            raise ValueError(
                f"all variables must have same shape, but variable {var_name!r} "
                f"has shape {shape!r}"
            )

        coords = var.coords
        if first_coords is None:
            first_coords = coords

        dask_chunks = var.chunks
        if dask_chunks:
            chunks = []
            for i in range(var.ndim):
                dim_name = var.dims[i]
                dim_chunk_sizes = dask_chunks[i]
                first_size = dim_chunk_sizes[0]
                if any(size != first_size for size in dim_chunk_sizes[1:-1]):
                    raise ValueError(
                        f"dimension {dim_name!r} of variable {var_name!r} has chunks of different sizes: "
                        f"{dim_chunk_sizes!r}"
                    )
                chunks.append(first_size)
            chunks = tuple(chunks)
            if first_chunks is None:
                first_chunks = chunks
            elif first_chunks != chunks:
                raise ValueError(
                    f"all variables must have same chunks, but variable {var_name!r} "
                    f"has chunks {chunks!r}"
                )

    if first_dims is None:
        raise ValueError("cube is empty")

    return CubeSchema(
        first_shape,
        first_coords,
        x_name=xy_var_names[0],
        y_name=xy_var_names[1],
        time_name=time_var_name,
        dims=tuple(str(d) for d in first_dims),
        chunks=first_chunks,
    )


def get_dataset_xy_var_names(
    coords: Union[xr.Dataset, xr.DataArray, Mapping[Hashable, xr.DataArray]],
    must_exist: bool = False,
    dataset_arg_name: str = "dataset",
) -> Optional[tuple[str, str]]:
    if hasattr(coords, "coords"):
        coords = coords.coords
    x_var_name = None
    y_var_name = None
    for var_name, var in coords.items():
        if (
            var.attrs.get("standard_name") == "projection_x_coordinate"
            or var.attrs.get("long_name") == "x coordinate of projection"
        ):
            if var.ndim == 1:
                x_var_name = var_name
        if (
            var.attrs.get("standard_name") == "projection_y_coordinate"
            or var.attrs.get("long_name") == "y coordinate of projection"
        ):
            if var.ndim == 1:
                y_var_name = var_name
        if x_var_name and y_var_name:
            return str(x_var_name), str(y_var_name)

    x_var_name = None
    y_var_name = None
    for var_name, var in coords.items():
        if var.attrs.get("long_name") == "longitude":
            if var.ndim == 1:
                x_var_name = var_name
        if var.attrs.get("long_name") == "latitude":
            if var.ndim == 1:
                y_var_name = var_name
        if x_var_name and y_var_name:
            return str(x_var_name), str(y_var_name)

    for x_var_name, y_var_name in (
        ("lon", "lat"),
        ("longitude", "latitude"),
        ("x", "y"),
    ):
        if x_var_name in coords and y_var_name in coords:
            x_var = coords[x_var_name]
            y_var = coords[y_var_name]
            if x_var.ndim == 1 and y_var.ndim == 1:
                return x_var_name, y_var_name

    if must_exist:
        raise ValueError(
            f"{dataset_arg_name} has no valid spatial coordinate variables"
        )


def get_dataset_time_var_name(
    dataset: Union[xr.Dataset, xr.DataArray],
    must_exist: bool = False,
    dataset_arg_name: str = "dataset",
) -> Optional[str]:
    time_var_name = "time"
    if time_var_name in dataset.coords:
        time_var = dataset.coords[time_var_name]
        if time_var.ndim == 1 and np.issubdtype(time_var.dtype, np.datetime64):
            return time_var_name

    if must_exist:
        raise ValueError(f"{dataset_arg_name} has no valid time coordinate variable")

    return None


def get_dataset_bounds_var_name(
    dataset: Union[xr.Dataset, xr.DataArray],
    var_name: str,
    must_exist: bool = False,
    dataset_arg_name: str = "dataset",
) -> Optional[str]:
    if var_name in dataset.coords:
        var = dataset[var_name]
        bounds_var_name = var.attrs.get("bounds", f"{var_name}_bnds")
        if bounds_var_name in dataset:
            bounds_var = dataset[bounds_var_name]
            if (
                bounds_var.ndim == 2
                and bounds_var.shape[0] == var.shape[0]
                and bounds_var.shape[1] == 2
            ):
                return bounds_var_name

    if must_exist:
        raise ValueError(
            f"{dataset_arg_name} has no valid bounds variable for variable {var_name!r}"
        )

    return None


def get_dataset_chunks(dataset: xr.Dataset) -> dict[Hashable, int]:
    """Get the most common chunk sizes for each
    chunked dimension of *dataset*.

    Note: Only data variables are considered.

    Args:
        dataset: A dataset.

    Returns:
        A dictionary that maps dimension names to common chunk sizes.
    """

    # Record the frequencies of chunk sizes for
    # each dimension d in each data variable var
    dim_size_counts: dict[Hashable, dict[int, int]] = {}
    for var_name, var in dataset.data_vars.items():
        if var.chunks:
            for d, c in zip(var.dims, var.chunks):
                # compute max chunk size max_c from
                # e.g.  c = (512, 512, 512, 193)
                max_c = max(0, *c)
                # for dimension d, save the frequencies
                # of the different max_c
                if d not in dim_size_counts:
                    size_counts = {max_c: 1}
                    dim_size_counts[d] = size_counts
                else:
                    size_counts = dim_size_counts[d]
                    if max_c not in size_counts:
                        size_counts[max_c] = 1
                    else:
                        size_counts[max_c] += 1

    # For each dimension d, determine the most frequently
    # seen chunk size max_c
    dim_sizes: dict[Hashable, int] = {}
    for d, size_counts in dim_size_counts.items():
        max_count = 0
        best_max_c = 0
        for max_c, count in size_counts.items():
            if count > max_count:
                # Should always come here, because count=1 is minimum
                max_count = count
                best_max_c = max_c
        assert best_max_c > 0
        dim_sizes[d] = best_max_c

    return dim_sizes


def rechunk_cube(
    cube: xr.Dataset,
    gm: GridMapping,
    chunks: Optional[dict[str, int]] = None,
    tile_size: Optional[tuple[int, int]] = None,
) -> tuple[xr.Dataset, GridMapping]:
    """Re-chunk data variables of *cube* so they all share the same chunk
    sizes for their dimensions.

    This functions rechunks *cube* for maximum compatibility with
    the Zarr format. Therefore it removes the "chunks" encoding
    from all variables.

    Args:
        cube: A data cube
        gm: The cube's grid mapping
        chunks: Optional mapping of dimension names to chunk sizes
        tile_size: Optional tile sizes, i.e. chunk size of spatial
            dimensions, given as (width, height)

    Returns:
        A potentially rechunked *cube* and adjusted grid mapping.
    """

    # get initial, common cube chunk sizes from given cube
    cube_chunks = get_dataset_chunks(cube)

    # Given chunks will overwrite initial values
    if chunks:
        for dim_name, size in chunks.items():
            cube_chunks[dim_name] = size

    # Given tile size will overwrite spatial dims
    x_dim_name, y_dim_name = gm.xy_dim_names
    if tile_size is not None:
        cube_chunks[x_dim_name] = tile_size[0]
        cube_chunks[y_dim_name] = tile_size[1]

    # Given grid mapping's tile size will overwrite
    # spatial dims only if missing still
    if gm.tile_size is not None:
        if x_dim_name not in cube_chunks:
            cube_chunks[x_dim_name] = gm.tile_size[0]
        if y_dim_name not in cube_chunks:
            cube_chunks[y_dim_name] = gm.tile_size[1]

    # If there is no chunking required, return identities
    if not cube_chunks:
        return cube, gm

    chunked_cube = xr.Dataset(attrs=cube.attrs)

    # Coordinate variables are always
    # chunked automatically
    chunked_cube = chunked_cube.assign_coords(
        coords={
            var_name: var.chunk({dim_name: "auto" for dim_name in var.dims})
            for var_name, var in cube.coords.items()
        }
    )

    # Data variables are chunked according to cube_chunks,
    # or if not specified, by the dimension size.
    chunked_cube = chunked_cube.assign(
        variables={
            var_name: var.chunk(
                {
                    dim_name: cube_chunks.get(dim_name, cube.sizes[dim_name])
                    for dim_name in var.dims
                }
            )
            for var_name, var in cube.data_vars.items()
        }
    )

    # Update chunks encoding for Zarr
    for var_name, var in chunked_cube.variables.items():
        if "chunks" in var.encoding:
            del var.encoding["chunks"]
        # if var.chunks is not None:
        #     # sizes[0] is the first of
        #     # e.g. sizes = (512, 512, 71)
        #     var.encoding.update(chunks=[
        #         sizes[0] for sizes in var.chunks
        #     ])
        # elif 'chunks' in var.encoding:
        #     del var.encoding['chunks']
        # print(f"--> {var_name}: encoding={var.encoding.get('chunks')!r}, chunks={var.chunks!r}")

    # Test whether tile size has changed after re-chunking.
    # If so, we will change the grid mapping too.
    tile_width = cube_chunks.get(x_dim_name)
    tile_height = cube_chunks.get(y_dim_name)
    assert tile_width is not None
    assert tile_height is not None
    tile_size = (tile_width, tile_height)
    if tile_size != gm.tile_size:
        # Note, changing grid mapping tile size may
        # rechunk (2D) coordinates in chunked_cube too
        gm = gm.derive(tile_size=tile_size)

    return chunked_cube, gm



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.


from .abc import MultiLevelDataset
from .base import BaseMultiLevelDataset
from .combined import CombinedMultiLevelDataset
from .computed import (
    ComputedMultiLevelDataset,
    augment_ml_dataset,
    open_ml_dataset_from_python_code,
)
from .fs import FsMultiLevelDataset, FsMultiLevelDatasetError
from .identity import IdentityMultiLevelDataset
from .lazy import LazyMultiLevelDataset
from .mapped import MappedMultiLevelDataset



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

DEFAULT_CALLABLE_NAME = "process_dataset"
DEFAULT_MODULE_NAME = "xcube_byoa"

TEMP_FILE_PREFIX = "xcube-byoa"



#  Copyright (c) 2018-2025 by xcube team and contributors
#  Permissions are hereby granted under the terms of the MIT License:
#  https://opensource.org/licenses/MIT.

from dataclasses import dataclass

from chartlets import Component


@dataclass(frozen=True)
class Markdown(Component):
    """A div-element in which the given markdown text is rendered."""

    text: str | None = None
    """The markdown text."""



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import warnings
from collections.abc import MutableMapping
from typing import Any, Dict

import click

from xcube.cli.common import (
    cli_option_dry_run,
    cli_option_quiet,
    cli_option_verbosity,
    configure_cli_output,
)
from xcube.constants import LOG

DELETE_ATTR_VALUE = "__delete__"


# noinspection PyShadowingBuiltins
@click.command(name="patch")
@click.argument("DATASET")
@click.option(
    "--metadata",
    "metadata_path",
    metavar="METADATA",
    help="The metadata to be patched."
    " Must be a JSON or YAML file"
    " using Zarr consolidated metadata format.",
)
@click.option(
    "--options",
    "options_path",
    metavar="OPTIONS",
    help="Protocol-specific storage options (see fsspec). Must be a JSON or YAML file.",
)
@cli_option_quiet
@cli_option_verbosity
@cli_option_dry_run
def patch(dataset, metadata_path, options_path, quiet, verbosity, dry_run):
    """
    Patch and consolidate the metadata of a dataset.

    DATASET can be either a local filesystem path or a URL.
    It must point to either a Zarr dataset (*.zarr)
    or a xcube multi-level dataset (*.levels).
    Additional storage options for a given protocol may be passed
    by the OPTIONS option.

    In METADATA, the special attribute value "__delete__" can be used to
    remove that attribute from dataset or array metadata.
    """
    configure_cli_output(quiet=quiet, verbosity=verbosity)
    metadata = load_metadata(metadata_path)
    storage_options = load_storage_options(options_path)
    patch_dataset(dataset, storage_options, metadata, dry_run)


def load_metadata(metadata_path):
    from xcube.util.config import load_json_or_yaml_config

    if not metadata_path:
        raise click.ClickException("Missing metadata to be patched")

    metadata = load_json_or_yaml_config(metadata_path)
    return parse_metadata(metadata)


def parse_metadata(metadata):
    if not isinstance(metadata, dict) or "zarr_consolidated_format" not in metadata:
        raise click.ClickException("Invalid consolidated metadata format")
    zarr_consolidated_format = metadata.get("zarr_consolidated_format")
    if zarr_consolidated_format != 1:
        raise click.ClickException("Unsupported consolidated metadata version")
    metadata = metadata.get("metadata")
    if not isinstance(metadata, dict):
        raise click.ClickException("Invalid metadata format")
    _metadata = dict()
    for k, v in metadata.items():
        if not isinstance(v, dict):
            raise click.ClickException(
                f'Invalid metadata format: entry "{k}" is not an object'
            )
        parts = k.split("/")
        if parts[-1] != ".zattrs":
            warnings.warn(f'Ignoring metadata entry "{k}": can only patch "*/.zattrs"')
        elif len(parts) not in (1, 2):
            warnings.warn(
                f'Ignoring metadata entry {k}: can only patch ".zattrs" of first level'
            )
        else:
            _metadata[k] = v
    if not _metadata:
        raise click.ClickException("No metadata provided")
    return _metadata


def load_storage_options(options_path):
    from xcube.util.config import load_json_or_yaml_config

    storage_options = {}
    if options_path:
        storage_options = load_json_or_yaml_config(options_path)
        if not isinstance(storage_options, dict):
            raise click.ClickException("Invalid storage options format")
    return storage_options


def patch_dataset(
    dataset_path: str,
    storage_options: dict[str, Any],
    metadata: dict[str, Any],
    dry_run: bool,
):
    if dataset_path.endswith(".levels"):
        fs, root = get_fs_and_root(dataset_path, storage_options)
        protocol = fs.protocol if isinstance(fs.protocol, str) else fs.protocol[0]
        prefix = f"{protocol}://" if protocol != "file" else ""
        for item in fs.listdir(root, detail=False):
            if item.endswith(".zarr"):
                _patch_dataset(prefix + item, storage_options, metadata, dry_run)
    else:
        _patch_dataset(dataset_path, storage_options, metadata, dry_run)


def _patch_dataset(
    dataset_path: str,
    storage_options: dict[str, Any],
    metadata: dict[str, Any],
    dry_run: bool,
):
    import zarr

    LOG.info(f"Opening {dataset_path}...")
    zarr_store = _open_zarr_store(dataset_path, storage_options)
    group = zarr.open(zarr_store, mode="r+" if not dry_run else "r")
    for k, v in metadata.items():
        parts = k.split("/")
        item = None
        if len(parts) == 1:
            item = group
        else:
            item_name, _ = parts
            if item_name in group:
                item = group[item_name]
            else:
                warnings.warn(f"Ignoring metadata entry {k}: not found in dataset")
        if item is not None:
            upd_count = 0
            del_count = 0
            for ak, av in v.items():
                if av == DELETE_ATTR_VALUE:
                    if ak in item.attrs:
                        if not dry_run:
                            del item.attrs[ak]
                        del_count += 1
                else:
                    if not dry_run:
                        item.attrs[ak] = av
                    upd_count += 1
            if upd_count and del_count:
                LOG.info(f"{k}: {upd_count} attribute(s) updated, {del_count} deleted")
            elif upd_count:
                LOG.info(f"{k}: {upd_count} attribute(s) updated")
            elif del_count:
                LOG.info(f"{k}: {del_count} attribute(s) deleted")

    if not dry_run:
        zarr.convenience.consolidate_metadata(zarr_store)
    LOG.info(f"Consolidated {dataset_path}")


def _open_zarr_store(
    dataset_path: str, storage_options: dict[str, Any]
) -> MutableMapping[str, bytes]:
    fs, root = get_fs_and_root(dataset_path, storage_options)
    return fs.get_mapper(root=root)


def get_fs_and_root(dataset_path: str, storage_options: dict[str, Any]):
    import fsspec

    protocol, root = fsspec.core.split_protocol(dataset_path)
    protocol = protocol or "file"
    fs = fsspec.filesystem(protocol, **storage_options)
    return fs, root



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Any, Callable, Dict, Optional

import jsonschema

from xcube.util.assertions import assert_instance
from xcube.util.jsonschema import JsonObjectSchema

from .error import DataStoreError


def assert_valid_params(
    params: Optional[dict[str, Any]],
    schema: Optional[JsonObjectSchema] = None,
    name: str = "params",
):
    """Assert that constructor/method parameters *params* are valid.

    Args:
        params: Dictionary of keyword arguments passed to a
            constructor/method.
        schema: The JSON Schema that *params* must adhere to.
        name: Name of the *params* variable.
    """
    _assert_valid(params, schema, name, "parameterization", _validate_params)


def assert_valid_config(
    config: Optional[dict[str, Any]],
    schema: Optional[JsonObjectSchema] = None,
    name: str = "config",
):
    """Assert that JSON object *config* is valid.

    Args:
        config: JSON object used for configuration.
        schema: The JSON Schema that *config* must adhere to.
        name: Name of the *config* variable.
    """
    _assert_valid(config, schema, name, "configuration", _validate_config)


def _validate_params(params: dict[str, Any], schema: JsonObjectSchema):
    # Note, params is a dictionary of Python objects.
    # Convert them to a JSON instance
    # and perform JSON Schema validation
    schema.validate_instance(params)


def _validate_config(config: dict[str, Any], schema: JsonObjectSchema):
    # Note, config is already a JSON object.
    # Perform JSON Schema validation directly.
    schema.validate_instance(config)


def _assert_valid(
    obj: Optional[dict[str, Any]],
    schema: Optional[JsonObjectSchema],
    name: str,
    kind: str,
    validator: Callable[[dict[str, Any], JsonObjectSchema], Any],
):
    if obj is None:
        return
    assert_instance(obj, dict, name=name)
    if schema is not None:
        assert_instance(schema, JsonObjectSchema, name=f"{name}_schema")
        try:
            validator(obj, schema)
        except jsonschema.ValidationError as e:
            raise DataStoreError(f"Invalid {kind} detected: {e.message}") from e



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import collections.abc
from collections.abc import Iterator
from typing import Dict, Set, Union

import xarray as xr

from xcube.core.mldataset import MultiLevelDataset

from ...util.assertions import assert_instance
from ..datasets.context import DatasetsContext

_LEVELS_EXT = ".levels"
_ZARR_EXT = ".zarr"


class DatasetsMapping(collections.abc.Mapping):
    """Represents the given *datasets_ctx* as a mapping from
    dataset identifier to dataset, it can
    be passed to class:ObjectStorage.

    This is the applied Adapter design pattern to make
    class:DatasetsContext compatible with the mapping argument for
    class:ObjectStorage.

    The original identifiers will be renamed in case their suffixes
    do not match the desired bucket type, so the bucket contents are
    more user-friendly. If *is_multi_level* is True, the new names
    will always have ".levels" suffix, otherwise the ".zarr" suffix.

    Args:
        datasets_ctx: The datasets' context
        is_multi_level: Whether this is a multi-level datasets' object
            storage
    """

    def __init__(self, datasets_ctx: DatasetsContext, is_multi_level: bool = False):
        assert_instance(datasets_ctx, DatasetsContext, name="datasets_ctx")
        assert_instance(is_multi_level, bool, name="is_multi_level")
        self._datasets_ctx = datasets_ctx
        self._is_multi_level = is_multi_level

    @property
    def _s3_names(self):
        return self._get_s3_names(self._datasets_ctx, self._is_multi_level)

    @staticmethod
    def _get_s3_names(
        datasets_ctx: DatasetsContext, is_multi_level: bool
    ) -> dict[str, str]:
        """Generate user-friendly S3 names for dataset identifiers.
        If *is_multi_level* is True, S3 names will be forced to have
        ".levels" suffix, otherwise the ".zarr" suffix.
        """
        ds_ids = [c["Identifier"] for c in datasets_ctx.get_dataset_configs()]
        all_ids = set(ds_ids)

        s3_names = {}
        for ds_id in ds_ids:
            s3_base, s3_ext = _split_base_ext(ds_id)
            if is_multi_level:
                s3_name = _replace_ext(s3_base, s3_ext, _ZARR_EXT, _LEVELS_EXT, all_ids)
            else:
                s3_name = _replace_ext(s3_base, s3_ext, _LEVELS_EXT, _ZARR_EXT, all_ids)
            s3_names[s3_name] = ds_id
        return s3_names

    def __len__(self) -> int:
        return len(self._s3_names)

    def __iter__(self) -> Iterator[str]:
        return iter(self._s3_names)

    def __contains__(self, s3_name: str) -> bool:
        """Check if *dataset_id* is a valid dataset.
        Overridden to avoid a call to __getitem__(),
        which will open the dataset (or raise ApiError!),
        but we want this to happen for direct __getitem__()
        calls only.
        """
        return s3_name in self._s3_names

    def __getitem__(self, s3_name: str) -> Union[xr.Dataset, MultiLevelDataset]:
        """Get or open the dataset given by *dataset_id*."""
        dataset_id = self._s3_names[s3_name]
        # Will raise ApiError
        if self._is_multi_level:
            return self._datasets_ctx.get_ml_dataset(dataset_id)
        else:
            return self._datasets_ctx.get_dataset(dataset_id)


def _split_base_ext(identifier: str):
    base_ext = identifier.rsplit(".", maxsplit=1)
    if len(base_ext) == 2:
        return base_ext[0], "." + base_ext[1]
    else:
        return identifier, ""


def _replace_ext(
    base_name: str, old_ext: str, trigger_ext: str, new_ext: str, all_ids: set[str]
) -> str:
    if old_ext == new_ext or (
        old_ext == trigger_ext and (base_name + new_ext) not in all_ids
    ):
        return base_name + new_ext
    return base_name + old_ext + new_ext



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.


__author__ = "Norman Fomferra (Brockmann Consult GmbH)"

UNDEFINED_STR = "UNDEFINED"


class _Undefined:
    """Represents the UNDEFINED value."""

    _hash_code = hash(UNDEFINED_STR) + 1

    def __str__(self):
        return UNDEFINED_STR

    def __repr__(self):
        return UNDEFINED_STR

    def __eq__(self, other):
        return self is other or isinstance(other, _Undefined)

    def __hash__(self) -> int:
        return _Undefined._hash_code


#: Singleton value used to indicate an undefined state.
UNDEFINED = _Undefined()



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.server.api import ApiHandler

from ...util.undefined import UNDEFINED
from ..datasets.routes import PATH_PARAM_DATASET_ID, PATH_PARAM_VAR_NAME
from .api import api
from .context import StatisticsContext
from .controllers import compute_statistics

QUERY_PARAM_X = {
    "name": "lon",
    "in": "query",
    "description": "Longitude in decimal degree",
    "required": True,
    "schema": {"type": "number", "minimum": -180, "maximum": 180},
}

QUERY_PARAM_Y = {
    "name": "lat",
    "in": "query",
    "description": "Latitude in decimal degree",
    "required": True,
    "schema": {"type": "number", "minimum": -90, "maximum": 90},
}

QUERY_PARAM_TIME = {
    "name": "time",
    "in": "query",
    "description": 'Timestamp using format "YYYY-MM-DD hh:mm:ss"',
    "required": False,
    "schema": {"type": "string", "format": "datetime"},
}


# noinspection PyPep8Naming
@api.route("/statistics/{datasetId}/{varName}")
class StatisticsHandler(ApiHandler[StatisticsContext]):
    @api.operation(
        operation_id="getValue",
        summary=(
            "Get the value of a dataset variable for given time stamp and location."
        ),
        parameters=[
            PATH_PARAM_DATASET_ID,
            PATH_PARAM_VAR_NAME,
            QUERY_PARAM_X,
            QUERY_PARAM_Y,
            QUERY_PARAM_TIME,
        ],
    )
    async def get(self, datasetId: str, varName: str):
        lon = self.request.get_query_arg("lon", type=float, default=UNDEFINED)
        lat = self.request.get_query_arg("lat", type=float, default=UNDEFINED)
        time = self.request.get_query_arg("time", type=str, default=None)
        trace_perf = self.request.get_query_arg(
            "debug", default=self.ctx.datasets_ctx.trace_perf
        )
        result = await self.ctx.run_in_executor(
            None,
            compute_statistics,
            self.ctx,
            datasetId,
            varName,
            (lon, lat),
            time,
            trace_perf,
        )
        await self.response.finish({"result": result})

    @api.operation(
        operation_id="getStatistics",
        summary=(
            "Get statistics of a dataset variable for given time stamp and geometry."
        ),
        description=(
            "The geometry is passed in the request body in"
            " form of a valid GeoJSON geometry object."
            " The operation returns the count, minimum, maximum, mean,"
            " and standard deviation of a data variable. If a 2-D geometry"
            " is passed, a histogram is returned as well."
        ),
        parameters=[
            PATH_PARAM_DATASET_ID,
            PATH_PARAM_VAR_NAME,
            QUERY_PARAM_TIME,
        ],
    )
    async def post(self, datasetId: str, varName: str):
        time = self.request.get_query_arg("time", type=str, default=None)
        trace_perf = self.request.get_query_arg(
            "debug", default=self.ctx.datasets_ctx.trace_perf
        )
        result = await self.ctx.run_in_executor(
            None,
            compute_statistics,
            self.ctx,
            datasetId,
            varName,
            self.request.json,
            time,
            trace_perf,
        )
        await self.response.finish({"result": result})



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import os
from abc import ABC, abstractmethod
from collections.abc import Iterator, Sequence
from typing import Any, Generic, List, Tuple, Type, TypeVar, Union

import geopandas
import xarray

from xcube.core.mldataset import MultiLevelDataset
from xcube.util.assertions import assert_instance
from xcube.util.jsonschema import JsonStringSchema

# A data type name or a DataType
DataTypeLike = Union[str, None, type, "DataType"]


class DataType:
    """
    A well-known Python data type that also has a representation
    using alias names.

    For example, the aliases "dataset" and "xarray.Dataset" both refer
    to the Python data type ``xarray.Dataset``.
    """

    _READTHEDOCS = os.environ.get("READTHEDOCS") == "True"
    _REGISTERED_DATA_TYPES: list["DataType"] = []

    @classmethod
    def register_data_type(cls, data_type: "DataType"):
        assert_instance(data_type, DataType, name="data_type")
        cls._REGISTERED_DATA_TYPES.append(data_type)

    def __init__(self, dtype: type, alias: Union[None, str, Sequence[str]] = None):
        """
        Args:
            dtype: The Python data type.
            alias: An alias name or list of aliases.
        """
        assert_instance(dtype, type if not self._READTHEDOCS else object, name="dtype")
        if alias is not None:
            assert_instance(alias, (str, tuple, list), name="alias")
        self._dtype = dtype
        self._aliases = (
            [] if alias is None else [alias] if isinstance(alias, str) else list(alias)
        ) + [self._get_fully_qualified_type_name(dtype)]
        self._alias_set = set(self._aliases)  # for faster lookup

    def __str__(self) -> str:
        return self.alias

    def __repr__(self) -> str:
        return repr(self.alias)

    def __eq__(self, other: Any):
        return self is other or (
            isinstance(other, DataType)
            and self.dtype is other.dtype
            and self.aliases == other.aliases
        )

    def __hash__(self):
        return hash(self.dtype) + 16 * hash(self.aliases)

    @property
    def dtype(self) -> type:
        """The Python data type."""
        return self._dtype

    @property
    def alias(self) -> str:
        """The primary alias name."""
        return self._aliases[0]

    @property
    def aliases(self) -> tuple[str, ...]:
        """All alias names."""
        return tuple(self._aliases)

    @classmethod
    def normalize(cls, data_type: DataTypeLike) -> "DataType":
        """Normalize the given *data_type* value into
        an instance of :class:`DataType`.

        Args:
            data_type: Another data type, maybe given as type alias name,
                as a type, or as a DataType instance.

        Returns: a DataType instance
        """
        if isinstance(data_type, DataType):
            return data_type
        if data_type is None or data_type is type(None):
            return ANY_TYPE
        if isinstance(data_type, str):
            for dt in cls._REGISTERED_DATA_TYPES:
                if data_type in dt._alias_set:
                    return dt
            raise ValueError(f"unknown data type {data_type!r}")
        if isinstance(data_type, type):
            for dt in cls._REGISTERED_DATA_TYPES:
                if data_type is dt.dtype:
                    return dt
            return DataType(data_type)
        raise ValueError(f"cannot convert {data_type!r} into a data type")

    def is_sub_type_of(self, data_type: DataTypeLike) -> bool:
        """Tests whether this data type is a sub-type of or the same
        as another data type.

        Args:
            data_type: The other data type, may be given as type
                alias name, as a type, or as a DataType instance.

        Returns: Whether this data type satisfies another data type.
        """
        return issubclass(self.dtype, self._normalize_dtype(data_type))

    def is_super_type_of(self, data_type: DataTypeLike) -> bool:
        """Tests whether this data type is a super-type of or the same
        as another data type.

        Args:
            data_type: The other data type, may be given as type
                alias name, as a type, or as a DataType instance.

        Returns: Whether this data type satisfies another data type.
        """
        return issubclass(self._normalize_dtype(data_type), self.dtype)

    @classmethod
    def get_schema(cls) -> JsonStringSchema:
        """Get the JSON Schema."""
        return JsonStringSchema(min_length=1, factory=cls.normalize, serializer=str)

    @classmethod
    def _normalize_dtype(cls, data_type: DataTypeLike) -> type:
        if isinstance(data_type, DataType):
            return data_type.dtype
        if isinstance(data_type, type):
            return data_type
        return cls.normalize(data_type).dtype

    @staticmethod
    def _get_fully_qualified_type_name(data_type: type) -> str:
        return f"{data_type.__module__}.{data_type.__name__}"


T = TypeVar("T")


class DataIterator(Generic[T], Iterator[T], ABC):
    """An iterator for data items.
    This class is a marker type for data stores that can return data iterators
    from their ``open_data()`` method.
    Data stores that support data iterators are not required to yield
    instances of this class. They just have to return iterator objects
    that yield instances of the desired item type, e.g. ``xarray.Dataset``.
    """

    @abstractmethod
    def __next__(self) -> T:
        """Yield the next data item."""


ANY_TYPE = DataType(object, ["any", "*", "object"])

DATASET_TYPE = DataType(xarray.Dataset, ["dataset", "xarray.Dataset"])

DATASET_ITERATOR_TYPE = DataType(
    DataIterator,  # Unfortunately we cannot use DataIterator[xarray.Dataset]
    ["dsiter", "DataIterator[xarray.Dataset]"],
)

MULTI_LEVEL_DATASET_TYPE = DataType(
    MultiLevelDataset,
    ["mldataset", "xcube.MultiLevelDataset", "xcube.core.mldataset.MultiLevelDataset"],
)

MULTI_LEVEL_DATASET_ITERATOR_TYPE = DataType(
    DataIterator,  # Unfortunately we cannot use DataIterator[MultiLevelDataset]
    ["mldsiter", "DataIterator[xcube.MultiLevelDataset]"],
)

GEO_DATA_FRAME_TYPE = DataType(
    geopandas.GeoDataFrame, ["geodataframe", "geopandas.GeoDataFrame"]
)

GEO_DATA_FRAME_ITERATOR_TYPE = DataType(
    DataIterator,  # Unfortunately we cannot use DataIterator[GeoDataFrame]
    ["gdfiter", "DataIterator[geopandas.GeoDataFrame]"],
)


def register_default_data_types():
    for data_type in [
        ANY_TYPE,
        DATASET_TYPE,
        DATASET_ITERATOR_TYPE,
        MULTI_LEVEL_DATASET_TYPE,
        MULTI_LEVEL_DATASET_ITERATOR_TYPE,
        GEO_DATA_FRAME_TYPE,
        GEO_DATA_FRAME_ITERATOR_TYPE,
    ]:
        DataType.register_data_type(data_type)


register_default_data_types()



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import os
from typing import Any

from xcube.util.versions import get_xcube_versions
from xcube.version import version

from .context import MetaContext


def get_service_info(ctx: MetaContext) -> dict[str, Any]:
    api_infos = []
    for other_api in ctx.apis:
        api_info = {
            "name": other_api.name,
            "version": other_api.version,
            "description": other_api.description,
        }
        api_infos.append({k: v for k, v in api_info.items() if v is not None})

    # TODO (forman): once APIs are configurable, we should
    #   get the server name, description from configuration too
    return dict(
        name="xcube Server",
        description="xcube Server",
        version=version,
        versions=get_xcube_versions(),
        apis=api_infos,
        startTime=ctx.start_time,
        currentTime=ctx.current_time,
        updateTime=ctx.update_time,
        pid=os.getpid(),
    )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import warnings
from collections.abc import Hashable, Mapping, Sequence
from typing import Any, Dict, Optional, Tuple, Union

import dask.array
import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr

from xcube.core.geom import get_dataset_bounds
from xcube.core.mldataset import MultiLevelDataset
from xcube.core.timecoord import (
    get_end_time_from_attrs,
    get_start_time_from_attrs,
    get_time_range_from_data,
    remove_time_part_from_isoformat,
)
from xcube.util.assertions import assert_given, assert_not_none, assert_true
from xcube.util.ipython import register_json_formatter
from xcube.util.jsonschema import (
    JsonArraySchema,
    JsonComplexSchema,
    JsonDateSchema,
    JsonDatetimeSchema,
    JsonIntegerSchema,
    JsonNumberSchema,
    JsonObject,
    JsonObjectSchema,
    JsonStringSchema,
)

from .datatype import (
    ANY_TYPE,
    DATASET_TYPE,
    GEO_DATA_FRAME_TYPE,
    MULTI_LEVEL_DATASET_TYPE,
    DataType,
    DataTypeLike,
)

# TODO: IMPORTANT: replace, reuse, or align with
#   xcube.core.schema.CubeSchema class
#   xcube.webapi.context.DatasetDescriptor type
#   responses of xcube.webapi.controllers.catalogue
# TODO: write tests
# TODO: validate params


def new_data_descriptor(
    data_id: str, data: Any, require: bool = False
) -> "DataDescriptor":
    if isinstance(data, MultiLevelDataset):
        dataset_descriptor_kwargs = _get_common_dataset_descriptor_props(
            data_id,
            # Note The highest level should have same metadata
            # and maybe loads faster.
            # data.get_dataset(data.num_levels - 1)
            data.get_dataset(0),
        )
        return MultiLevelDatasetDescriptor(
            num_levels=data.num_levels, **dataset_descriptor_kwargs
        )

    if isinstance(data, xr.Dataset):
        dataset_descriptor_kwargs = _get_common_dataset_descriptor_props(data_id, data)
        return DatasetDescriptor(**dataset_descriptor_kwargs)

    if isinstance(data, gpd.GeoDataFrame):
        # TODO: implement me: data -> GeoDataFrameDescriptor
        return GeoDataFrameDescriptor(data_id=data_id)

    if not require:
        return DataDescriptor(data_id=data_id, data_type=ANY_TYPE)

    raise NotImplementedError()


def _get_common_dataset_descriptor_props(
    data_id: str, dataset: Union[xr.Dataset, MultiLevelDataset]
) -> dict[str, Any]:
    dims = {str(k): v for k, v in dataset.sizes.items()}
    coords = _build_variable_descriptor_dict(dataset.coords)
    data_vars = _build_variable_descriptor_dict(dataset.data_vars)
    spatial_res = _determine_spatial_res(dataset)
    bbox = _determine_bbox(dataset)
    time_range = _determine_time_coverage(dataset)
    time_period = _determine_time_period(dataset)
    return dict(
        data_id=data_id,
        dims=dims,
        coords=coords,
        data_vars=data_vars,
        bbox=bbox,
        time_range=time_range,
        time_period=time_period,
        spatial_res=spatial_res,
        attrs=dataset.attrs,
    )


class DataDescriptor(JsonObject):
    """A generic descriptor for any data.
    Also serves as a base class for more specific data descriptors.

    Args:
        data_id: An identifier for the data
        data_type: A type specifier for the data
        crs: A coordinate reference system identifier, as an EPSG, PROJ
            or WKT string
        bbox: A bounding box of the data
        time_range: Start and end time delimiting this data's temporal
            extent;
        time_period: The data's periodicity if it is evenly temporally
            resolved.
        open_params_schema: A JSON schema describing the parameters that
            may be used to open this data.
    """

    def __init__(
        self,
        data_id: str,
        data_type: DataTypeLike,
        *,
        crs: str = None,
        bbox: tuple[float, float, float, float] = None,
        time_range: tuple[Optional[str], Optional[str]] = None,
        time_period: str = None,
        open_params_schema: JsonObjectSchema = None,
        **additional_properties,
    ):
        assert_given(data_id, "data_id")
        if additional_properties:
            warnings.warn(
                f"Additional properties received;"
                f" will be ignored: {additional_properties}"
            )
        self.data_id = data_id
        self.data_type = DataType.normalize(data_type)
        self.crs = crs
        self.bbox = tuple(bbox) if bbox else None
        self.time_range = tuple(time_range) if time_range else None
        self.time_period = time_period
        self.open_params_schema = open_params_schema

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                data_id=JsonStringSchema(min_length=1),
                data_type=DataType.get_schema(),
                crs=JsonStringSchema(min_length=1),
                bbox=JsonArraySchema(
                    items=[
                        JsonNumberSchema(),
                        JsonNumberSchema(),
                        JsonNumberSchema(),
                        JsonNumberSchema(),
                    ]
                ),
                time_range=JsonComplexSchema(
                    any_of=[
                        JsonDateSchema.new_range(nullable=True),
                        JsonDatetimeSchema.new_range(nullable=True),
                    ]
                ),
                time_period=JsonStringSchema(min_length=1),
                open_params_schema=JsonObjectSchema(additional_properties=True),
            ),
            required=["data_id", "data_type"],
            additional_properties=True,
            factory=cls,
        )


class DatasetDescriptor(DataDescriptor):
    """A descriptor for a gridded, N-dimensional dataset represented
    by xarray.Dataset. Comprises a description of the data variables
    contained in the dataset.

    Regrading *time_range* and *time_period* parameters, please refer to
    https://github.com/dcs4cop/xcube/blob/main/docs/source/storeconv.md#date-time-and-duration-specifications

    Args:
        data_id: An identifier for the data
        data_type: The data type of the data described
        crs: A coordinate reference system identifier, as an EPSG, PROJ
            or WKT string
        bbox: A bounding box of the data
        time_range: Start and end time delimiting this data's temporal
            extent
        time_period: The data's periodicity if it is evenly temporally
            resolved
        spatial_res: The spatial extent of a pixel in crs units
        dims: A mapping of the dataset's dimensions to their sizes
        coords: mapping of the dataset's data coordinate names to
            instances of :class:`VariableDescriptor`
        data_vars: A mapping of the dataset's variable names to
            instances of :class:`VariableDescriptor`
        attrs: A mapping containing arbitrary attributes of the dataset
        open_params_schema: A JSON schema describing the parameters that
            may be used to open this data
    """

    def __init__(
        self,
        data_id: str,
        *,
        data_type: DataTypeLike = DATASET_TYPE,
        crs: str = None,
        bbox: tuple[float, float, float, float] = None,
        time_range: tuple[Optional[str], Optional[str]] = None,
        time_period: str = None,
        spatial_res: float = None,
        dims: Mapping[str, int] = None,
        coords: Mapping[str, "VariableDescriptor"] = None,
        data_vars: Mapping[str, "VariableDescriptor"] = None,
        attrs: Mapping[Hashable, any] = None,
        open_params_schema: JsonObjectSchema = None,
        **additional_properties,
    ):
        super().__init__(
            data_id=data_id,
            data_type=data_type,
            crs=crs,
            bbox=bbox,
            time_range=time_range,
            time_period=time_period,
            open_params_schema=open_params_schema,
        )
        assert_true(
            DATASET_TYPE.is_super_type_of(data_type)
            or MULTI_LEVEL_DATASET_TYPE.is_super_type_of(data_type),
            f"illegal data_type,"
            f" must be compatible with {DATASET_TYPE!r}"
            f" or {MULTI_LEVEL_DATASET_TYPE!r}",
        )
        if additional_properties:
            warnings.warn(
                f"Additional properties received;"
                f" will be ignored: {additional_properties}"
            )
        self.dims = dict(dims) if dims else None
        self.spatial_res = spatial_res
        self.coords = coords if coords else None
        self.data_vars = data_vars if data_vars else None
        self.attrs = _attrs_to_json(attrs) if attrs else None

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        schema = super().get_schema()
        schema.properties.update(
            dims=JsonObjectSchema(additional_properties=JsonIntegerSchema(minimum=0)),
            spatial_res=JsonNumberSchema(exclusive_minimum=0.0),
            coords=JsonObjectSchema(
                additional_properties=VariableDescriptor.get_schema()
            ),
            data_vars=JsonObjectSchema(
                additional_properties=VariableDescriptor.get_schema()
            ),
            attrs=JsonObjectSchema(additional_properties=True),
        )
        schema.required = ["data_id", "data_type"]
        schema.additional_properties = False
        schema.factory = cls
        return schema


class VariableDescriptor(JsonObject):
    """A descriptor for dataset variable represented by
    xarray.DataArray instances.
    They are part of dataset descriptor for an gridded, N-dimensional
    dataset represented by
    xarray.Dataset.

    Args:
        name: The variable name
        dtype: The data type of the variable.
        dims: A list of the names of the variable's dimensions.
        chunks: A list of the chunk sizes of the variable's dimensions
        attrs: A mapping containing arbitrary attributes of the variable
    """

    def __init__(
        self,
        name: str,
        dtype: str,
        dims: Sequence[str],
        *,
        chunks: Sequence[int] = None,
        attrs: Mapping[Hashable, any] = None,
        **additional_properties,
    ):
        assert_given(name, "name")
        assert_given(dtype, "dtype")
        assert_not_none(dims, "dims")
        if additional_properties:
            warnings.warn(
                f"Additional properties received;"
                f" will be ignored: {additional_properties}"
            )
        self.name = name
        self.dtype = dtype
        self.dims = tuple(dims)
        self.chunks = tuple(chunks) if chunks else None
        self.attrs = _attrs_to_json(attrs) if attrs else None

    @property
    def ndim(self) -> int:
        """Number of dimensions."""
        return len(self.dims)

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                name=JsonStringSchema(min_length=1),
                dtype=JsonStringSchema(min_length=1),
                dims=JsonArraySchema(items=JsonStringSchema(min_length=1)),
                chunks=JsonArraySchema(items=JsonIntegerSchema(minimum=0)),
                attrs=JsonObjectSchema(additional_properties=True),
            ),
            required=["name", "dtype", "dims"],
            additional_properties=False,
            factory=cls,
        )


class MultiLevelDatasetDescriptor(DatasetDescriptor):
    """A descriptor for a gridded, N-dimensional, multi-level,
    multi-resolution dataset represented by
    xcube.core.mldataset.MultiLevelDataset.

    Args:
        data_id: An identifier of the multi-level dataset
        num_levels: The number of levels of this multi-level dataset
        data_type: A type specifier for the multi-level dataset
    """

    def __init__(
        self,
        data_id: str,
        num_levels: int,
        *,
        data_type: DataTypeLike = MULTI_LEVEL_DATASET_TYPE,
        **kwargs,
    ):
        assert_given(data_id, "data_id")
        assert_given(num_levels, "num_levels")
        super().__init__(data_id=data_id, data_type=data_type, **kwargs)
        assert_true(
            MULTI_LEVEL_DATASET_TYPE.is_super_type_of(data_type),
            f"illegal data_type, must be compatible with {MULTI_LEVEL_DATASET_TYPE!r}",
        )
        self.num_levels = num_levels

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        schema = super().get_schema()
        schema.properties.update(
            num_levels=JsonIntegerSchema(minimum=1),
        )
        schema.required.append("num_levels")
        schema.additional_properties = False
        schema.factory = cls
        return schema


class GeoDataFrameDescriptor(DataDescriptor):
    """A descriptor for a geo-vector dataset represented by a
    geopandas.GeoDataFrame instance.

    Args:
        data_id: An identifier of the geopandas.GeoDataFrame
        feature_schema: A schema describing the properties of the vector
            data
        kwargs: Parameters passed to super :class:`DataDescriptor`
    """

    def __init__(
        self,
        data_id: str,
        *,
        data_type: DataTypeLike = GEO_DATA_FRAME_TYPE,
        feature_schema: JsonObjectSchema = None,
        **kwargs,
    ):
        super().__init__(data_id=data_id, data_type=data_type, **kwargs)
        assert_true(
            GEO_DATA_FRAME_TYPE.is_super_type_of(data_type),
            f"illegal data_type, must be compatible with {GEO_DATA_FRAME_TYPE!r}",
        )
        self.feature_schema = feature_schema

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        schema = super().get_schema()
        schema.properties.update(
            feature_schema=JsonObjectSchema(additional_properties=True),
        )
        schema.required = ["data_id"]
        schema.additional_properties = False
        schema.factory = cls
        return schema


register_json_formatter(DataDescriptor)
register_json_formatter(DatasetDescriptor)
register_json_formatter(VariableDescriptor)
register_json_formatter(MultiLevelDatasetDescriptor)
register_json_formatter(GeoDataFrameDescriptor)


#############################################################################
# Implementation helpers


def _build_variable_descriptor_dict(variables) -> Mapping[str, "VariableDescriptor"]:
    return {
        str(var_name): VariableDescriptor(
            name=str(var_name),
            dtype=str(var.dtype),
            dims=var.dims,
            chunks=(
                tuple([max(chunk) for chunk in tuple(var.chunks)])
                if var.chunks
                else None
            ),
            attrs=var.attrs,
        )
        for var_name, var in variables.items()
    }


def _determine_bbox(data: xr.Dataset) -> Optional[tuple[float, float, float, float]]:
    try:
        return get_dataset_bounds(data)
    except ValueError:
        if (
            "geospatial_lon_min" in data.attrs
            and "geospatial_lat_min" in data.attrs
            and "geospatial_lon_max" in data.attrs
            and "geospatial_lat_max" in data.attrs
        ):
            return (
                data.geospatial_lon_min,
                data.geospatial_lat_min,
                data.geospatial_lon_max,
                data.geospatial_lat_max,
            )


def _determine_spatial_res(data: xr.Dataset):
    # TODO get rid of these hard-coded coord names as soon as
    #   new resampling is available
    lat_dimensions = ["lat", "latitude", "y"]
    for lat_dimension in lat_dimensions:
        if lat_dimension in data:
            lat_diff = data[lat_dimension].diff(dim=data[lat_dimension].dims[0]).values
            lat_res = lat_diff[0]
            lat_regular = np.allclose(lat_res, lat_diff, 1e-8)
            if lat_regular:
                return float(abs(lat_res))


def _determine_time_coverage(data: xr.Dataset):
    start_time, end_time = get_time_range_from_data(data)
    if start_time is not None:
        try:
            start_time = remove_time_part_from_isoformat(
                pd.to_datetime(start_time).isoformat()
            )
        except TypeError:
            start_time = None
    if start_time is None:
        start_time = get_start_time_from_attrs(data)
    if end_time is not None:
        try:
            end_time = remove_time_part_from_isoformat(
                pd.to_datetime(end_time).isoformat()
            )
        except TypeError:
            end_time = None
    if end_time is None:
        end_time = get_end_time_from_attrs(data)
    return start_time, end_time


def _determine_time_period(data: xr.Dataset):
    if "time" in data and len(data["time"].values) > 1:
        time_diff = (
            data["time"].diff(dim=data["time"].dims[0]).values.astype(np.float64)
        )
        time_res = time_diff[0]
        time_regular = np.allclose(time_res, time_diff, 1e-8)
        if time_regular:
            time_period = pd.to_timedelta(time_res).isoformat()
            # remove leading P
            time_period = time_period[1:]
            # removing sub-day precision
            return time_period.split("T")[0]


def _attrs_to_json(attrs: Mapping[Hashable, Any]) -> Optional[dict[str, Any]]:
    new_attrs: dict[str, Any] = {}
    for k, v in attrs.items():
        if isinstance(v, np.ndarray):
            v = v.tolist()
        elif isinstance(v, dask.array.Array):
            v = np.array(v).tolist()
        if isinstance(v, float) and np.isnan(v):
            v = None
        new_attrs[str(k)] = v
    return new_attrs



#  Copyright (c) 2018-2025 by xcube team and contributors
#  Permissions are hereby granted under the terms of the MIT License:
#  https://opensource.org/licenses/MIT.

from .markdown import Markdown

__all__ = [
    "Markdown",
]



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from abc import ABC, abstractmethod
from collections.abc import Iterator

from xcube.util.jsonschema import JsonObjectSchema

from .assertions import assert_valid_params
from .datatype import DataTypeLike
from .descriptor import DataDescriptor


class DataSearcher(ABC):
    """Allow searching data in a data store."""

    @classmethod
    @abstractmethod
    def get_search_params_schema(
        cls, data_type: DataTypeLike = None
    ) -> JsonObjectSchema:
        """Get the schema for the parameters that can be passed
        as *search_params* to :meth:`search_data`.
        Parameters are named and described by the properties of the
        returned JSON object schema.

        Args:
            data_type: If given, the search parameters will be tailored
                to search for data for the given *data_type*.

        Returns:
            A JSON object schema whose properties describe this store's
            search parameters.
        """

    @abstractmethod
    def search_data(
        self, data_type: DataTypeLike = None, **search_params
    ) -> Iterator[DataDescriptor]:
        """Search this store for data resources.
        If *data_type* is given, the search is restricted
        to data resources of that type.

        Returns an iterator over the search results which are
        returned as :class:`DataDescriptor` objects.

        If a store implementation supports only a single data type,
        it should verify that *data_type*
        is either None or compatible with the supported data type
        specifier.

        Args:
            data_type: An optional data type that is known to be
                supported by this data store.
            **search_params: The search parameters.

        Returns:
            An iterator of data descriptors for the found data
            resources.

        Raises:
            DataStoreError: If an error occurs.
        """


# noinspection PyUnresolvedReferences
class DefaultSearchMixin(DataSearcher):
    """A mixin for data store implementations that implements default
    search behaviour.

    It is expected that such data stores have no dedicated search
    parameters.
    """

    # noinspection PyUnusedLocal
    @classmethod
    def get_search_params_schema(
        cls, data_type: DataTypeLike = None
    ) -> JsonObjectSchema:
        """Get search parameters JSON object schema.

        The default implementation returns a schema that does
        not allow for any search parameters.

        Args:
            data_type

        Returns:
            a JSON object schema for the search parameters
        """
        return JsonObjectSchema(additional_properties=False)

    def search_data(
        self, data_type: DataTypeLike = None, **search_params
    ) -> Iterator[DataDescriptor]:
        """Search the data store.

        The default implementation returns all data resources that
        may be filtered using the optional *data_type*.

        Args:
            data_type: Type specifier to filter returned data resources.
            **search_params: Not supported (yet)

        Returns:
            an iterator of :class:`DataDescriptor` instances
        """
        search_params_schema = self.get_search_params_schema(data_type=data_type)
        assert_valid_params(
            search_params, name="search_params", schema=search_params_schema
        )
        for data_id in self.get_data_ids(data_type=data_type):
            data_descriptor = self.describe_data(data_id, data_type=data_type)
            yield data_descriptor



#  Copyright (c) 2018-2025 by xcube team and contributors
#  Permissions are hereby granted under the terms of the MIT License:
#  https://opensource.org/licenses/MIT.

import xarray as xr


def compute_indexes(dataset: xr.Dataset) -> xr.Dataset:
    b04 = dataset.B04
    b05 = dataset.B05
    b06 = dataset.B06
    b08 = dataset.B08
    b11 = dataset.B11
    scl = dataset.SCL

    # See https://gisgeography.com/sentinel-2-bands-combinations/

    # moisture_index
    #
    moisture_index = (b08 - b11) / (b08 + b11)
    moisture_index.attrs.update(
        color_value_min=0.0,
        color_value_max=0.5,
        color_bar_name="Blues",
        units="-",
        description="Simple moisture index: (B08-B11) / (B08+B11)",
    )
    # where valid and not cloud
    moisture_index = moisture_index.where((scl >= 2) & (scl <= 7))

    # vegetation_index
    #
    vegetation_index = (b08 - b04) / (b08 + b04)
    vegetation_index.attrs.update(
        color_value_min=0.0,
        color_value_max=0.25,
        color_bar_name="Greens",
        units="-",
        description="Simple vegetation index or NDVI: (B08-B04) / (B08+B04)",
    )
    # where water
    vegetation_index = vegetation_index.where(scl == 6)

    # chlorophyll_index
    #
    b_from, b_peek, b_to = b04, b05, b06
    wlen_from = b04.attrs["wavelength"]
    wlen_peek = b05.attrs["wavelength"]
    wlen_to = b06.attrs["wavelength"]
    f = (wlen_peek - wlen_from) / (wlen_to - wlen_from)
    chlorophyll_index = (b_peek - b_from) - f * (b_to - b_from)
    chlorophyll_index.attrs.update(
        color_value_min=0.0,
        color_value_max=0.025,
        color_bar_name="viridis",
        units="-",
        description="Maximum chlorophyll index: (B05-B04) - f * (B06-B04)",
    )
    # where water
    chlorophyll_index = chlorophyll_index.where(scl == 6)

    return xr.Dataset(
        dict(
            moisture_index=moisture_index,
            vegetation_index=vegetation_index,
            chlorophyll_index=chlorophyll_index,
        )
    )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import click

DEFAULT_OUTPUT_PATH = "out.csv"
DEFAULT_NUM_POINTS_MAX = 100


@click.command(name="genpts", hidden=True)
@click.argument("cube")
@click.option(
    "--output",
    "-o",
    "output_path",
    help=f'Output path. Defaults to "{DEFAULT_OUTPUT_PATH}".',
    default=DEFAULT_OUTPUT_PATH,
)
@click.option(
    "--number",
    "-n",
    "num_points_max",
    help=f'Number of point_data. Defaults to "{DEFAULT_NUM_POINTS_MAX}".',
    type=int,
    default=DEFAULT_NUM_POINTS_MAX,
)
def genpts(cube: str, output_path: str, num_points_max: int):
    """Generate synthetic data points from CUBE.

    Generates synthetic data points for a given data CUBE and write points to a CSV and GeoJSON file.
    The primary use of the tool is to point datasets for machine learning tasks
    and to create test points for the "xcube extract" command.
    """
    import numpy as np
    import pandas as pd
    import xarray as xr

    def _all_data_vars_valid(dataset: xr.DataArray):
        for _, var in dataset.data_vars.items():
            if not np.isfinite(var):
                return False
        return True

    cube = xr.open_zarr(cube)

    num_points = 0
    point_data = dict(time=[], lat=[], lon=[])
    for var_name in cube.data_vars:
        point_data[var_name] = []

    spatial_res = ((cube.lon[-1] - cube.lon[0]) / (cube.lon.size - 1)).values
    temporal_res = ((cube.time[-1] - cube.time[0]) // (cube.time.size - 1)).values

    while num_points < num_points_max:
        it = np.random.randint(0, cube.time.size)
        iy = np.random.randint(0, cube.lat.size)
        ix = np.random.randint(0, cube.lon.size)

        point = cube.isel(time=it, lat=iy, lon=ix)

        if _all_data_vars_valid(point):
            point_data["time"].append(
                point.time.values + 0.5 * temporal_res * np.random.logistic(scale=0.2)
            )
            point_data["lat"].append(
                point.lat.values + 0.5 * spatial_res * np.random.logistic(scale=0.2)
            )
            point_data["lon"].append(
                point.lon.values + 0.5 * spatial_res * np.random.logistic(scale=0.2)
            )
            for var_name, var in point.data_vars.items():
                value = var.values
                if np.issubdtype(value.dtype, np.floating):
                    value += np.random.logistic(scale=0.05)
                point_data[var_name].append(value)
            num_points += 1

    if output_path.endswith(".csv"):
        pd.DataFrame(point_data).to_csv(output_path)
    elif output_path.endswith(".geojson"):
        import json

        with open(output_path, "w") as fp:
            json.dump(_to_geojson_dict(point_data, num_points), fp, indent=2)


def _to_geojson_dict(point_data, num_points):
    import numpy as np

    features = []
    for i in range(num_points):
        x = float(point_data["lon"][i])
        y = float(point_data["lat"][i])
        properties = {}
        for k, v in point_data.items():
            if k in {"lon", "lat"}:
                continue
            value = v[i]
            if np.issubdtype(value.dtype, np.floating):
                value = float(value)
            elif np.issubdtype(value.dtype, np.datetime64):
                value = str(value)
            else:
                value = int(value)
            properties[k] = value
        features.append(
            {
                "type": "Feature",
                "id": i,
                "geometry": {
                    "type": "Point",
                    "coordinates": (x, y),
                },
                "properties": properties,
            }
        )
    return {"type": "FeatureCollection", "features": features}


if __name__ == "__main__":
    genpts.main()



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import xarray as xr


def dump_dataset(dataset: xr.Dataset, var_names=None, show_var_encoding=False) -> str:
    """Dumps a dataset or variables into a text string.

    Args:
        dataset: input dataset
        var_names: names of variables to be dumped
        show_var_encoding: also dump variable encodings?

    Returns:
        the dataset dump
    """
    lines = []
    if not var_names:
        lines.append(str(dataset))
        if show_var_encoding:
            for var_name, var in dataset.coords.items():
                if var.encoding:
                    lines.append(
                        dump_var_encoding(
                            var,
                            header=f"Encoding for coordinate variable {var_name!r}:",
                        )
                    )
            for var_name, var in dataset.data_vars.items():
                if var.encoding:
                    lines.append(
                        dump_var_encoding(
                            var, header=f"Encoding for data variable {var_name!r}:"
                        )
                    )
    else:
        for var_name in var_names:
            var = dataset[var_name]
            lines.append(str(var))
            if show_var_encoding and var.encoding:
                lines.append(dump_var_encoding(var))
    return "\n".join(lines)


def dump_var_encoding(var: xr.DataArray, header="Encoding:", indent=4) -> str:
    """Dump the encoding information of a variable into a text string.

    Args:
        var: Dataset variable.
        header: Title/header string.
        indent: Indention in spaces.

    Returns:
        the variable dump
    """
    lines = [header]
    max_len = 0
    for k in var.encoding:
        max_len = max(max_len, len(k))
    indent_spaces = indent * " "
    for k, v in var.encoding.items():
        tab_spaces = (2 + max_len - len(k)) * " "
        lines.append(f"{indent_spaces}{k}:{tab_spaces}{v!r}")
    return "\n".join(lines)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import collections.abc
import json
import math
import pathlib
import warnings
from collections.abc import Mapping, Sequence
from functools import cached_property
from typing import Any, Dict, List, Optional, Union

import fsspec
import fsspec.core
import numpy as np
import xarray as xr
import zarr

# noinspection PyUnresolvedReferences
import xcube.core.zarrstore
from xcube.core.gridmapping import GridMapping
from xcube.core.subsampling import AggMethod, AggMethods
from xcube.util.assertions import assert_instance
from xcube.util.fspath import get_fs_path_class, resolve_path
from xcube.util.types import ScalarOrPair, normalize_scalar_or_pair

from .abc import MultiLevelDataset
from .base import BaseMultiLevelDataset
from .lazy import LazyMultiLevelDataset

LEVELS_FORMAT_VERSION = "1.0"


class FsMultiLevelDataset(LazyMultiLevelDataset):
    _MIN_CACHE_SIZE = 1024 * 1024  # 1 MiB

    def __init__(
        self,
        path: str,
        fs: Optional[fsspec.AbstractFileSystem] = None,
        fs_root: Optional[str] = None,
        fs_kwargs: Optional[Mapping[str, Any]] = None,
        cache_size: Optional[int] = None,
        consolidate: Optional[bool] = None,
        **zarr_kwargs,
    ):
        if fs is None:
            fs, path = fsspec.core.url_to_fs(path, **(fs_kwargs or {}))
        assert_instance(fs, fsspec.AbstractFileSystem, name="fs")
        assert_instance(path, str, name="data_id")
        # TODO: Setting ds_id=path is likely the root cause for
        #   https://github.com/xcube-dev/xcube/issues/1007.
        #   Then, the actual fix is to replace slashes by dashes or underscores.
        #   But this requires deeper investigation and more test cases.
        #   A quick fix has been applied in `xcube.webapi.viewer.Viewer.add_dataset()`
        #   implementation. forman, 2024-06-07
        super().__init__(ds_id=path)
        self._path = path
        self._fs = fs
        self._fs_root = fs_root
        self._cache_size = cache_size
        self._consolidate = consolidate
        self._zarr_kwargs = zarr_kwargs
        self._path_class = get_fs_path_class(fs)

    @property
    def path(self) -> str:
        return self._path

    @property
    def fs(self) -> fsspec.AbstractFileSystem:
        return self._fs

    @property
    def cache_size(self) -> Optional[int]:
        return self._cache_size

    @cached_property
    def size_weights(self) -> np.ndarray:
        """Size weights are used to distribute the cache size
        over the levels.
        """
        return self.compute_size_weights(self.num_levels)

    @cached_property
    def tile_size(self) -> Optional[Sequence[int]]:
        spec = self._levels_spec
        return spec.get("tile_size")

    @cached_property
    def use_saved_levels(self) -> Optional[bool]:
        spec = self._levels_spec
        return spec.get("use_saved_levels")

    @cached_property
    def base_dataset_path(self) -> Optional[str]:
        spec = self._levels_spec
        return spec.get("base_dataset_path")

    @cached_property
    def base_dataset_params(self) -> Optional[dict[str, Any]]:
        spec = self._levels_spec
        return spec.get("base_dataset_params")

    @cached_property
    def agg_methods(self) -> Optional[Mapping[str, AggMethod]]:
        spec = self._levels_spec
        return spec.get("agg_methods")

    @cached_property
    def _levels_spec(self) -> Mapping[str, Any]:
        path = f"{self._path}/.zlevels"
        spec = {}
        if self.fs.exists(path):
            with self.fs.open(f"{self._path}/.zlevels") as fp:
                spec = json.load(fp)
            if not isinstance(spec, collections.abc.Mapping):
                raise TypeError("Unexpected .zlevels file. Must be a JSON object.")
        # TODO (forman): validate JSON object
        return spec

    def _get_dataset_lazily(self, index: int, parameters) -> xr.Dataset:
        cache_size = self._cache_size

        fs = self._fs

        open_params = dict(self._zarr_kwargs)
        base_dataset_open_params = None

        ds_path = self._get_path(self._path)
        link_path = ds_path / f"{index}.link"
        if fs.isfile(str(link_path)):
            # If file "{index}.link" exists, we have a link to
            # a level Zarr and open this instead,
            with fs.open(str(link_path), "r") as fp:
                level_path = self._get_path(fp.read())
            if not level_path.is_absolute() and not self._is_path_relative_to_path(
                level_path, ds_path
            ):
                level_path = resolve_path(ds_path / level_path)
            base_dataset_open_params = self.base_dataset_params
        else:
            # Nominal "{index}.zarr" must exist
            level_path = ds_path / f"{index}.zarr"

        if isinstance(base_dataset_open_params, dict):
            # TODO: complete logic here
            engine = base_dataset_open_params.pop("engine", "zarr")

        level_zarr_store = fs.get_mapper(str(level_path))

        consolidated = (
            self._consolidate
            if self._consolidate is not None
            else (".zmetadata" in level_zarr_store)
        )

        if isinstance(cache_size, int) and cache_size >= self._MIN_CACHE_SIZE:
            # compute cache size for level weighted by
            # size in pixels for each level
            cache_size = math.ceil(self.size_weights[index] * cache_size)
            if cache_size >= self._MIN_CACHE_SIZE:
                level_zarr_store = zarr.LRUStoreCache(
                    level_zarr_store, max_size=cache_size
                )

        try:
            level_dataset = xr.open_zarr(
                level_zarr_store, consolidated=consolidated, **self._zarr_kwargs
            )
        except ValueError as e:
            raise FsMultiLevelDatasetError(
                f"Failed to open dataset {level_path!r}: {e}"
            ) from e

        level_dataset.zarr_store.set(level_zarr_store)
        return level_dataset

    @staticmethod
    def _is_path_relative_to_path(level_path, ds_path):
        if hasattr(level_path, "is_relative_to"):
            # Python >=3.9
            return level_path.is_relative_to(ds_path)
        try:
            # Python <3.9
            level_path.relative_to(ds_path)
            return True
        except ValueError:
            return False

    @classmethod
    def compute_size_weights(cls, num_levels: int) -> np.ndarray:
        weights = (2 ** np.arange(0, num_levels, dtype=np.float64)) ** 2
        return weights[::-1] / np.sum(weights)

    def _get_num_levels_lazily(self) -> int:
        spec = self._levels_spec
        num_levels = spec.get("num_levels")
        levels = self._get_levels()
        if num_levels is None:
            num_levels = len(levels)
        expected_levels = list(range(num_levels))
        for level in expected_levels:
            if level != levels[level]:
                raise FsMultiLevelDatasetError(
                    f"Inconsistent"
                    f" multi-level dataset {self.ds_id!r},"
                    f" expected levels {expected_levels!r}"
                    f" found {levels!r}"
                )
        return num_levels

    def _get_levels(self) -> list[int]:
        levels = []
        paths = [
            self._get_path(entry["name"])
            for entry in self._fs.listdir(self._path, detail=True)
        ]
        for path in paths:
            # No ext, i.e. dir_name = "<level>", is proposed by
            # https://github.com/zarr-developers/zarr-specs/issues/50.
            # xcube already selected dir_name = "<level>.zarr".
            basename = path.stem
            if path.stem and path.suffix in ("", ".zarr", ".link"):
                try:
                    level = int(basename)
                except ValueError:
                    continue
                levels.append(level)
        levels = sorted(levels)
        return levels

    def _get_path(self, *args) -> pathlib.PurePath:
        return self._path_class(*args)

    @classmethod
    def write_dataset(
        cls,
        dataset: Union[xr.Dataset, MultiLevelDataset],
        path: str,
        fs: Optional[fsspec.AbstractFileSystem] = None,
        fs_root: Optional[str] = None,
        fs_kwargs: Optional[Mapping[str, Any]] = None,
        replace: bool = False,
        num_levels: Optional[int] = None,
        consolidated: bool = True,
        tile_size: Optional[ScalarOrPair[int]] = None,
        use_saved_levels: bool = False,
        base_dataset_path: Optional[str] = None,
        base_dataset_params: Optional[dict[str, Any]] = None,
        agg_methods: Optional[AggMethods] = None,
        **zarr_kwargs,
    ) -> str:
        assert_instance(dataset, (xr.Dataset, MultiLevelDataset), name="dataset")

        if fs is None:
            fs, path = fsspec.core.url_to_fs(path, **(fs_kwargs or {}))
        if tile_size is not None:
            tile_size = normalize_scalar_or_pair(
                tile_size, item_type=int, name="tile_size"
            )

        assert_instance(path, str, name="path")
        assert_instance(fs, fsspec.AbstractFileSystem, name="fs")

        if isinstance(dataset, MultiLevelDataset):
            ml_dataset = dataset
            if tile_size:
                warnings.warn("tile_size is ignored for multi-level datasets")
            if agg_methods:
                warnings.warn("agg_methods is ignored for multi-level datasets")
        else:
            base_dataset: xr.Dataset = dataset
            grid_mapping = None
            if tile_size is not None:
                grid_mapping = GridMapping.from_dataset(base_dataset)
                x_name, y_name = grid_mapping.xy_dim_names
                # noinspection PyTypeChecker
                base_dataset = base_dataset.chunk(
                    {x_name: tile_size[0], y_name: tile_size[1]}
                )
                # noinspection PyTypeChecker
                grid_mapping = grid_mapping.derive(tile_size=tile_size)
            ml_dataset = BaseMultiLevelDataset(
                base_dataset,
                grid_mapping=grid_mapping,
                num_levels=num_levels,
                agg_methods=agg_methods,
            )

        if use_saved_levels:
            ml_dataset = BaseMultiLevelDataset(
                ml_dataset.get_dataset(0),
                grid_mapping=ml_dataset.grid_mapping,
                agg_methods=agg_methods,
            )

        path_class = get_fs_path_class(fs)
        data_path = path_class(path)
        fs.mkdirs(str(data_path), exist_ok=replace)

        if num_levels is None or num_levels <= 0:
            num_levels_max = ml_dataset.num_levels
        else:
            num_levels_max = min(num_levels, ml_dataset.num_levels)

        with fs.open(str(data_path / ".zlevels"), mode="w") as fp:
            levels_data: dict[str, Any] = dict(
                version=LEVELS_FORMAT_VERSION, num_levels=num_levels_max
            )
            if use_saved_levels is not None:
                levels_data.update(use_saved_levels=bool(use_saved_levels))
            if base_dataset_path:
                levels_data.update(base_dataset_path=base_dataset_path)
            if base_dataset_params is not None:
                levels_data.update(base_dataset_params=base_dataset_params)
            if tile_size is not None:
                levels_data.update(tile_size=list(tile_size))
            if hasattr(ml_dataset, "agg_methods"):
                levels_data.update(agg_methods=dict(ml_dataset.agg_methods))
            json.dump(levels_data, fp, indent=2)

        for index in range(num_levels_max):
            level_dataset = ml_dataset.get_dataset(index)
            if base_dataset_path and index == 0:
                assert_instance(fs_root, str, name="fs_root")

                # Write file "0.link" instead of copying
                # level zero dataset to "0.zarr".

                # Compute a relative base dataset path first
                base_dataset_path = path_class(fs_root, base_dataset_path)
                data_parent_path = data_path.parent
                try:
                    base_dataset_path = base_dataset_path.relative_to(data_parent_path)
                except ValueError as e:
                    raise FsMultiLevelDatasetError(
                        f"Invalid base_dataset_id: {base_dataset_path}"
                    ) from e
                base_dataset_path = ".." / base_dataset_path

                # Then write relative base dataset path into link file
                link_path = data_path / f"{index}.link"
                with fs.open(str(link_path), mode="w") as fp:
                    fp.write(base_dataset_path.as_posix())
            else:
                # Write level "{index}.zarr"
                level_path = data_path / f"{index}.zarr"
                level_zarr_store = fs.get_mapper(str(level_path), create=True)
                try:
                    level_dataset.to_zarr(
                        level_zarr_store,
                        mode="w" if replace else None,
                        consolidated=consolidated,
                        **zarr_kwargs,
                    )
                except ValueError as e:
                    # TODO: remove already written data!
                    raise FsMultiLevelDatasetError(
                        f"Failed to write dataset {path}: {e}"
                    ) from e
                if use_saved_levels:
                    level_dataset = xr.open_zarr(
                        level_zarr_store, consolidated=consolidated
                    )
                    level_dataset.zarr_store.set(level_zarr_store)
                    ml_dataset.set_dataset(index, level_dataset)

        return path


class FsMultiLevelDatasetError(ValueError):
    def __init__(self, message: str):
        super().__init__(message)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Any, Callable

import numpy as np
import xarray as xr

from .exprvar import ExprVar


def get_constants() -> dict[str, Any]:
    return {
        "nan": np.nan,
        "e": np.e,
        "inf": np.inf,
        "pi": np.pi,
    }


def get_numpy_ufuncs() -> dict[str, np.ufunc]:
    """Get numpy universal functions (considered safe)"""
    # noinspection PyProtectedMember
    return {
        k: v
        for k, v in np.__dict__.items()
        if isinstance(v, np.ufunc) and isinstance(k, str) and not k.startswith("_")
    }


def get_xarray_funcs() -> dict[str, Callable]:
    """Get safe xarray functions"""
    # noinspection PyProtectedMember
    return {"where": xr.where}


# noinspection PyProtectedMember
_GLOBAL_NAMES = {
    **get_constants(),
    **{k: ExprVar._wrap_fn(fn) for k, fn in get_numpy_ufuncs().items()},
    **{k: ExprVar._wrap_fn(fn) for k, fn in get_xarray_funcs().items()},
}

# noinspection PyProtectedMember
del ExprVar._wrap_fn



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.server.api import Context
from xcube.webapi.common.context import ResourcesContext

from ..datasets.context import DatasetsContext
from .dsmapping import DatasetsMapping
from .objectstorage import ObjectStorage


class S3Context(ResourcesContext):
    """Context for S3 API."""

    def __init__(self, server_ctx: Context):
        super().__init__(server_ctx)
        self._datasets_ctx = server_ctx.get_api_ctx("datasets")
        self._buckets = {
            "datasets": ObjectStorage(DatasetsMapping(self._datasets_ctx, False)),
            "pyramids": ObjectStorage(
                DatasetsMapping(self._datasets_ctx, True),
            ),
        }

    @property
    def datasets_ctx(self) -> DatasetsContext:
        return self._datasets_ctx

    def get_bucket(self, bucket_name: str) -> ObjectStorage:
        return self._buckets[bucket_name]



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import sys

import click

from xcube.cli.common import (
    cli_option_scheduler,
    cli_option_traceback,
    configure_logging,
    configure_warnings,
    handle_cli_exception,
    new_cli_ctx_obj,
)
from xcube.constants import EXTENSION_POINT_CLI_COMMANDS, LOG_LEVEL_OFF_NAME, LOG_LEVELS
from xcube.util.plugin import get_extension_registry
from xcube.version import version


# noinspection PyShadowingBuiltins,PyUnusedLocal
@click.group(name="xcube")
@click.version_option(version)
@cli_option_traceback
@cli_option_scheduler
@click.option(
    "--loglevel",
    "log_level",
    metavar="LOG_LEVEL",
    type=click.Choice(LOG_LEVELS),
    default=LOG_LEVEL_OFF_NAME,
    help=f"Log level."
    f" Must be one of {', '.join(LOG_LEVELS)}."
    f" Defaults to {LOG_LEVEL_OFF_NAME}."
    f" If the level is not {LOG_LEVEL_OFF_NAME},"
    f" any log messages up to the given level will be"
    f" written either to the console (stderr)"
    f" or LOG_FILE, if provided.",
)
@click.option(
    "--logfile",
    "log_file",
    metavar="LOG_FILE",
    help=f"Log file path."
    f" If given, any log messages will redirected into"
    f" LOG_FILE. Disables console output"
    f" unless otherwise enabled, e.g.,"
    f" using the --verbose flag."
    f" Effective only if LOG_LEVEL"
    f" is not {LOG_LEVEL_OFF_NAME}.",
)
@click.option(
    "--warnings",
    "-w",
    is_flag=True,
    help="Show any warnings emitted during operation (warnings are hidden by default).",
)
def cli(traceback=False, scheduler=None, log_level=None, log_file=None, warnings=None):
    """xcube Toolkit"""
    configure_logging(log_file=log_file, log_level=log_level)
    configure_warnings(warnings)


_cli_commands_registered = False


def _register_cli_commands():
    global _cli_commands_registered
    if _cli_commands_registered:
        return
    _cli_commands_registered = True
    # Add registered CLI commands
    registry = get_extension_registry()
    for command in registry.find_components(EXTENSION_POINT_CLI_COMMANDS):
        cli.add_command(command)


_register_cli_commands()


def main(args=None):
    # noinspection PyBroadException
    ctx_obj = new_cli_ctx_obj()
    try:
        exit_code = cli.main(args=args, obj=ctx_obj, standalone_mode=False)
    except Exception as e:
        exit_code = handle_cli_exception(
            e, traceback_mode=ctx_obj.get("traceback", False)
        )
    sys.exit(exit_code)


if __name__ == "__main__":
    main()



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import abc
from collections.abc import Awaitable
from typing import Any, Callable, Optional, TypeVar, Union

from tornado import concurrent

ReturnT = TypeVar("ReturnT")


class AsyncExecution(abc.ABC):
    """
    An interface that defines asynchronous function execution.
    """

    @abc.abstractmethod
    def call_later(
        self, delay: Union[int, float], callback: Callable, *args, **kwargs
    ) -> object:
        """Executes the given callable *callback* after *delay* seconds.

        The method returns a handle that can be used to cancel the
        callback.

        Args:
            delay: Delay in seconds.
            callback: Callback to be called.
            args: Positional arguments passed to *callback*.
            kwargs: Keyword arguments passed to *callback*.

        Returns: A handle that provides the methods
            ``cancel()`` and ``cancelled()``.
        """

    @abc.abstractmethod
    def run_in_executor(
        self,
        executor: Optional[concurrent.futures.Executor],
        function: Callable[..., ReturnT],
        *args: Any,
        **kwargs: Any,
    ) -> Awaitable[ReturnT]:
        """Concurrently runs a *function* in a
        ``concurrent.futures.Executor``.
        If *executor* is ``None``, the framework's default
        executor will be used.

        Args:
            executor: An optional executor.
            function: The function to be run concurrently.
            args: Positional arguments passed to *function*.
            kwargs: Keyword arguments passed to *function*.

        Returns: The awaitable return value of *function*.
        """



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from abc import ABC, abstractmethod
from typing import Any, Optional

import xarray as xr

from xcube.constants import EXTENSION_POINT_DATA_OPENERS, EXTENSION_POINT_DATA_WRITERS
from xcube.util.assertions import assert_given
from xcube.util.extension import Extension, ExtensionPredicate, ExtensionRegistry
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.plugin import get_extension_registry

from .datatype import DataType, DataTypeLike
from .error import DataStoreError
from .preload import PreloadHandle

#######################################################
# Data accessor instantiation and registry query
#######################################################


def new_data_opener(
    opener_id: str,
    extension_registry: Optional[ExtensionRegistry] = None,
    **opener_params,
) -> "DataOpener":
    """Get an instance of the data opener identified by *opener_id*.

    The optional, extra opener parameters *opener_params* may
    be used by data store (``xcube.core.store.DataStore``)
    implementations so they can share their internal state with the opener.

    Args:
        opener_id: The data opener identifier.
        extension_registry: Optional extension registry. If not given,
            the global extension registry will be used.
        **opener_params: Extra opener parameters.

    Returns:
        A data opener instance.
    """
    assert_given(opener_id, "opener_id")
    extension_registry = extension_registry or get_extension_registry()
    if not extension_registry.has_extension(EXTENSION_POINT_DATA_OPENERS, opener_id):
        raise DataStoreError(f"A data opener named {opener_id!r} is not registered")
    return extension_registry.get_component(EXTENSION_POINT_DATA_OPENERS, opener_id)(
        **opener_params
    )


def new_data_writer(
    writer_id: str,
    extension_registry: Optional[ExtensionRegistry] = None,
    **writer_params,
) -> "DataWriter":
    """Get an instance of the data writer identified by *writer_id*.

    The optional, extra writer parameters *writer_params* may be used by
    data store (``xcube.core.store.DataStore``) implementations so they
    can share their internal state with the writer.

    Args:
        writer_id: The data writer identifier.
        extension_registry: Optional extension registry. If not given,
            the global extension registry will be used.
        **writer_params: Extra writer parameters.

    Returns:
        A data writer instance.
    """
    assert_given(writer_id, "writer_id")
    extension_registry = extension_registry or get_extension_registry()
    if not extension_registry.has_extension(EXTENSION_POINT_DATA_WRITERS, writer_id):
        raise DataStoreError(f"A data writer named {writer_id!r} is not registered")
    return extension_registry.get_component(EXTENSION_POINT_DATA_WRITERS, writer_id)(
        **writer_params
    )


def find_data_opener_extensions(
    predicate: ExtensionPredicate = None,
    extension_registry: Optional[ExtensionRegistry] = None,
) -> list[Extension]:
    """Get registered data opener extensions using the optional
    filter function *predicate*.

    Args:
        predicate: An optional filter function.
        extension_registry: Optional extension registry. If not given,
            the global extension registry will be used.

    Returns:
        List of matching extensions.
    """
    extension_registry = extension_registry or get_extension_registry()
    return extension_registry.find_extensions(
        EXTENSION_POINT_DATA_OPENERS, predicate=predicate
    )


def find_data_writer_extensions(
    predicate: ExtensionPredicate = None,
    extension_registry: Optional[ExtensionRegistry] = None,
) -> list[Extension]:
    """Get registered data writer extensions using the optional filter
    function *predicate*.

    Args:
        predicate: An optional filter function.
        extension_registry: Optional extension registry. If not given,
            the global extension registry will be used.

    Returns:
        List of matching extensions.
    """
    extension_registry = extension_registry or get_extension_registry()
    return extension_registry.find_extensions(
        EXTENSION_POINT_DATA_WRITERS, predicate=predicate
    )


def get_data_accessor_predicate(
    data_type: DataTypeLike = None, format_id: str = None, storage_id: str = None
) -> ExtensionPredicate:
    """Get a predicate that checks if the name of a data accessor extension is
    compliant with *data_type*, *format_id*, *storage_id*.

    Args:
        data_type: Optional data data type to be supported. May be given
            as type alias name, as a type, or as a DataType instance.
        format_id: Optional data format identifier to be supported.
        storage_id: Optional data storage identifier to be supported.

    Returns:
        A filter function.

    Raises:
        DataStoreError: If an error occurs.
    """
    if any((data_type, format_id, storage_id)):
        data_type = DataType.normalize(data_type) if data_type is not None else None

        def _predicate(extension: Extension) -> bool:
            extension_parts = extension.name.split(":", maxsplit=4)
            if storage_id is not None:
                ext_storage_id = extension_parts[2]
                if ext_storage_id != "*" and ext_storage_id != storage_id:
                    return False
            if format_id is not None:
                ext_format_id = extension_parts[1]
                if ext_format_id != "*" and ext_format_id != format_id:
                    return False
            if data_type is not None:
                ext_data_type = DataType.normalize(extension_parts[0])
                if not data_type.is_super_type_of(ext_data_type):
                    return False
            return True

    else:
        # noinspection PyUnusedLocal
        def _predicate(extension: Extension) -> bool:
            return True

    return _predicate


#######################################################
# Interface Classes
#######################################################


class DataOpener(ABC):
    """An interface that specifies a parameterized `open_data()` operation.

    Possible open parameters are implementation-specific and
    are described by a JSON Schema.

    Note this interface uses the term "opener" to underline the expected
    laziness of the operation. For example, when a xarray.Dataset is
    returned from a Zarr directory, the actual data is represented by
    Dask arrays and will be loaded only on-demand.
    """

    @abstractmethod
    def get_open_data_params_schema(self, data_id: str = None) -> JsonObjectSchema:
        """Get the schema for the parameters passed as *open_params* to
        :meth:`open_data`.
        If *data_id* is given, the returned schema will be tailored
        to the constraints implied by the identified data resource.
        Some openers might not support this, therefore *data_id*
        is optional, and if it is omitted, the returned schema will be
        less restrictive.

        Args:
            data_id: An optional data resource identifier.

        Returns:
            The schema for the parameters in *open_params*.

        Raises:
            DataStoreError: If an error occurs.
        """

    @abstractmethod
    def open_data(self, data_id: str, **open_params) -> Any:
        """Open the data resource given by the data resource identifier
        *data_id* using the supplied *open_params*.

        Raises if *data_id* does not exist.

        Args:
            data_id: The data resource identifier.
            **open_params: Opener-specific parameters.

        Returns:
            An xarray.Dataset instance.

        Raises:
            DataStoreError: If an error occurs.
        """

    def close(self):
        """Closes this data opener.

        Should be called if the data opener is no longer needed.

        This method may close local files, remote connections, or release
        allocated resources.

        The default implementation does nothing.
        """

    def __del__(self):
        """Overridden to call ``close()``."""
        self.close()


class DataDeleter(ABC):
    """An interface that specifies a parameterized `delete_data()` operation.

    Possible delete parameters are implementation-specific and
    are described by a JSON Schema.
    """

    @abstractmethod
    def get_delete_data_params_schema(self, data_id: str = None) -> JsonObjectSchema:
        """Get the schema for the parameters passed as *delete_params*
        to :meth:`delete_data`.
        If *data_id* is given, the returned schema will be tailored to
        the constraints implied by the identified data resource.
        Some deleters might not support this, therefore *data_id*
        is optional, and if it is omitted, the returned schema will
        be less restrictive.

        Args:
            data_id: An optional data resource identifier.

        Returns:
            The schema for the parameters in *delete_params*.

        Raises:
            DataStoreError: If an error occurs.
        """

    @abstractmethod
    def delete_data(self, data_id: str, **delete_params):
        """Delete a data resource. Raises if *data_id* does not exist.

        Args:
            data_id: A data resource identifier known to exist.
            **delete_params: Deleter-specific parameters.

        Raises:
            DataStoreError: If an error occurs.
        """


class DataWriter(DataDeleter, ABC):
    """An interface that specifies a parameterized `write_data()` operation.

    Possible write parameters are implementation-specific and
    are described by a JSON Schema.
    """

    @abstractmethod
    def get_write_data_params_schema(self) -> JsonObjectSchema:
        """Get the schema for the parameters passed as *write_params* to
        :meth:`write_data`.

        Returns:
            The schema for the parameters in *write_params*.

        Raises:
            DataStoreError: If an error occurs.
        """

    @abstractmethod
    def write_data(
        self, data: Any, data_id: str, replace: bool = False, **write_params
    ) -> str:
        """Write a data resource using the supplied *data_id* and *write_params*.

        Args:
            data: The data resource's in-memory representation to be
                written.
            data_id: A unique data resource identifier.
            replace: Whether to replace an existing data resource.
            **write_params: Writer-specific parameters.

        Returns:
            The data resource identifier used to write the data
            resource.

        Raises:
            DataStoreError: If an error occurs.
        """


class DataPreloader(ABC):
    """An interface that specifies a parameterized `preload_data()` operation.

        Warning: This is an experimental and potentially unstable API
        introduced in xcube 1.8.

        Many data store implementations rely on remote data APIs.
        Such API may provide only limited data access performance.
        Hence, the approach taken by ``DataStore.open_data(data_id, ...)`` alone
        is suboptimal for a user's perspective. This is because the method is
        blocking as it is not asynchronous, it may take long time before it
        returns, and it cannot report any progress while doing so.
        The reasons for slow and unresponsive data APIs are manifold: intended
        access is by file download, access is bandwidth limited, or not allowing
        for sub-setting.

    Data stores may implement the ``preload_data()`` method differently—or not at all.
    In most cases, if preloading is required, the data will be downloaded and stored
    temporarily in a cache for access.
    """

    @abstractmethod
    def preload_data(
        self,
        *data_ids: str,
        **preload_params: Any,
    ) -> PreloadHandle:
        """Preload the given data items for faster access.

        Warning: This is an experimental and potentially unstable API
        introduced in xcube 1.8.

        The method may be blocking or non-blocking.
        Implementations may offer the following keyword arguments
        in *preload_params*:

        - ``blocking``: whether the preload process is blocking.
          Should be `True` by default if supported.
        - ``monitor``: a callback function that serves as a progress monitor.
          It receives the preload handle and the recent partial state update.

        Args:
            data_ids: Data identifiers to be preloaded.
            preload_params: data store specific preload parameters.
              See method ``get_preload_data_params_schema()`` for information
              on the possible options.

        Returns:
            A handle for the preload process. The default implementation
            returns an empty preload handle.
        """

    # noinspection PyMethodMayBeStatic
    @abstractmethod
    def get_preload_data_params_schema(self) -> JsonObjectSchema:
        """Get the JSON schema that describes the keyword
        arguments that can be passed to ``preload_data()``.

        Returns:
            A ``JsonObjectSchema`` object whose properties describe
            the parameters of ``preload_data()``.
        """


class DataTimeSliceUpdater(DataWriter, ABC):
    """An interface that specifies writing of time slice data."""

    @abstractmethod
    def append_data_time_slice(self, data_id: str, time_slice: xr.Dataset):
        """Append a time slice to the identified data resource.

        Args:
            data_id: The data resource identifier.
            time_slice: The time slice data to be inserted. Must be
                compatible with the data resource.

        Raises:
            DataStoreError: If an error occurs.
        """

    @abstractmethod
    def insert_data_time_slice(self, data_id: str, time_slice: Any, time_index: int):
        """Insert a time slice into the identified data resource at given index.

        Args:
            data_id: The data resource identifier.
            time_slice: The time slice data to be inserted. Must be
                compatible with the data resource.
            time_index: The time index.

        Raises:
            DataStoreError: If an error occurs.
        """

    @abstractmethod
    def replace_data_time_slice(self, data_id: str, time_slice: Any, time_index: int):
        """Replace a time slice in the identified data resource at given index.

        Args:
            data_id: The data resource identifier.
            time_slice: The time slice data to be inserted. Must be
                compatible with the data resource.
            time_index: The time index.

        Raises:
            DataStoreError: If an error occurs.
        """



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import pandas as pd

from xcube.server.api import ApiHandler

from ..datasets import PATH_PARAM_DATASET_ID, PATH_PARAM_VAR_NAME
from .api import api
from .context import TimeSeriesContext
from .controllers import get_time_series


# noinspection PyPep8Naming
@api.route("/timeseries/{datasetId}/{varName}")
class TimeseriesHandler(ApiHandler[TimeSeriesContext]):
    @api.operation(
        operation_id="getTimeSeries",
        summary="Get the time-series for a variable and given GeoJSON object.",
        parameters=[
            PATH_PARAM_DATASET_ID,
            PATH_PARAM_VAR_NAME,
            {
                "name": "aggMethods",
                "in": "query",
                "description": "Comma-separated list of"
                " aggregation methods."
                " Valid methods are"
                ' "mean", "median", "std",'
                ' "min", "max", "count"',
                "schema": {
                    "type": "string",
                },
            },
            {
                "name": "startDate",
                "in": "query",
                "description": "Start timestamp",
                "schema": {"type": "string", "format": "datetime"},
            },
            {
                "name": "endDate",
                "in": "query",
                "description": "End timestamp",
                "schema": {"type": "string", "format": "datetime"},
            },
            {
                "name": "tolerance",
                "in": "query",
                "description": "Time tolerance in seconds that"
                " expands the given time range",
                "schema": {
                    "type": "string",
                    "default": 1.0,
                },
            },
            {
                "name": "maxValids",
                "in": "query",
                "description": "Maximum number of valid"
                " time-series values"
                " to be returned.",
                "schema": {
                    "type": "integer",
                },
            },
        ],
    )
    async def post(self, datasetId: str, varName: str):
        geo_json_object = self.request.json
        agg_methods = self.request.get_query_arg("aggMethods", type=str, default=None)
        agg_methods = agg_methods.split(",") if agg_methods else None
        start_date = self.request.get_query_arg(
            "startDate", type=pd.Timestamp, default=None
        )
        end_date = self.request.get_query_arg(
            "endDate", type=pd.Timestamp, default=None
        )
        tolerance = self.request.get_query_arg("tolerance", type=float, default=1.0)
        max_valids = self.request.get_query_arg("maxValids", type=int, default=None)
        result = await self.ctx.run_in_executor(
            None,
            get_time_series,
            self.ctx,
            datasetId,
            varName,
            geo_json_object,
            agg_methods,
            start_date,
            end_date,
            tolerance,
            max_valids,
        )
        self.response.set_header("Content-Type", "application/json")
        await self.response.finish(dict(result=result))



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.server.api import Api
from xcube.webapi.ows.coverages.context import CoveragesContext

api = Api(
    "ows.coverages",
    description="OGC API - Coverages",
    required_apis=["datasets"],
    create_ctx=CoveragesContext,
    config_schema=None,
)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Union

import xarray as xr

from xcube.core.mldataset import (
    FsMultiLevelDataset,
    FsMultiLevelDatasetError,
    MultiLevelDataset,
)
from xcube.core.subsampling import AGG_METHODS

# Note, we need the following reference to register the
# xarray property accessor
# noinspection PyUnresolvedReferences
from xcube.core.zarrstore import ZarrStoreHolder
from xcube.util.assertions import assert_instance
from xcube.util.jsonschema import (
    JsonArraySchema,
    JsonBooleanSchema,
    JsonComplexSchema,
    JsonIntegerSchema,
    JsonNullSchema,
    JsonObjectSchema,
    JsonStringSchema,
)

from ... import DataStoreError
from ...datatype import DATASET_TYPE, MULTI_LEVEL_DATASET_TYPE, DataType
from .dataset import DatasetZarrFsDataAccessor


class MultiLevelDatasetLevelsFsDataAccessor(DatasetZarrFsDataAccessor):
    """Opener/writer extension name 'mldataset:levels:<protocol>'."""

    @classmethod
    def get_data_type(cls) -> DataType:
        return MULTI_LEVEL_DATASET_TYPE

    @classmethod
    def get_format_id(cls) -> str:
        return "levels"

    def open_data(self, data_id: str, **open_params) -> MultiLevelDataset:
        assert_instance(data_id, str, name="data_id")
        fs, fs_root, open_params = self.load_fs(open_params)
        try:
            return FsMultiLevelDataset(data_id, fs=fs, fs_root=fs_root, **open_params)
        except FsMultiLevelDatasetError as e:
            raise DataStoreError(f"{e}") from e

    def get_write_data_params_schema(self) -> JsonObjectSchema:
        schema = super().get_write_data_params_schema()  # creates deep copy
        # TODO: remove use_saved_levels, instead see #619
        schema.properties["use_saved_levels"] = JsonBooleanSchema(
            description="Whether to open an already saved level"
            " and downscale it then."
            " May be used to avoid computation of"
            " entire Dask graphs at each level.",
            default=False,
        )
        schema.properties["base_dataset_id"] = JsonStringSchema(
            description="If given, avoids writing the base dataset"
            ' at level 0. Instead a file "{data_id}/0.link"'
            " is created whose content is the given base dataset"
            " identifier.",
            nullable=True,
        )
        schema.properties["base_dataset_open_params"] = JsonObjectSchema(
            description=(
                "Optional parameters used for opening the base dataset"
                " at level 0 if `base_dataset_id` is provided."
            ),
            additional_properties=True,
            nullable=True,
        )
        schema.properties["tile_size"] = JsonComplexSchema(
            one_of=[
                JsonIntegerSchema(minimum=1),
                JsonArraySchema(
                    items=[JsonIntegerSchema(minimum=1), JsonIntegerSchema(minimum=1)]
                ),
                JsonNullSchema(),
            ],
            description="Tile size to be used for all levels of the"
            " written multi-level dataset.",
        )
        schema.properties["num_levels"] = JsonIntegerSchema(
            description=(
                "Upper limit for the number of levels to be written."
                " The actual number may be lower if an additional level would"
                " result in a tile size below the allowed threshold given"
                " by the `tile_size` argument, if any."
            ),
            minimum=1,
            nullable=True,
        )
        schema.properties["agg_methods"] = JsonComplexSchema(
            one_of=[
                JsonStringSchema(enum=AGG_METHODS),
                JsonObjectSchema(
                    additional_properties=JsonStringSchema(enum=AGG_METHODS)
                ),
                JsonNullSchema(),
            ],
            description="Aggregation method for the pyramid levels."
            ' If not explicitly set, "first" is used for integer '
            ' variables and "mean" for floating point variables.'
            " If given as object, it is a mapping from variable "
            " name pattern to aggregation method. The pattern"
            " may include wildcard characters * and ?.",
        )
        return schema

    def write_data(
        self,
        data: MultiLevelDataset,
        data_id: str,
        replace: bool = False,
        **write_params,
    ) -> str:
        assert_instance(data, MultiLevelDataset, name="data")
        return self.write_data_generic(data, data_id, replace=replace, **write_params)

    def write_data_generic(
        self,
        data: Union[xr.Dataset, MultiLevelDataset],
        data_id: str,
        replace: bool = False,
        **write_params,
    ) -> str:
        assert_instance(data, (xr.Dataset, MultiLevelDataset), name="data")
        fs, fs_root, write_params = self.load_fs(write_params)
        base_dataset_id = write_params.pop("base_dataset_id", None)
        try:
            return FsMultiLevelDataset.write_dataset(
                data,
                data_id,
                fs=fs,
                fs_root=fs_root,
                replace=replace,
                base_dataset_path=base_dataset_id,
                **write_params,
            )
        except FsMultiLevelDatasetError as e:
            raise DataStoreError(f"{e}") from e


class DatasetLevelsFsDataAccessor(MultiLevelDatasetLevelsFsDataAccessor):
    """Opener/writer extension name 'dataset:levels:<protocol>'."""

    @classmethod
    def get_data_type(cls) -> DataType:
        return DATASET_TYPE

    def open_data(self, data_id: str, **open_params) -> xr.Dataset:
        ml_dataset = super().open_data(data_id, **open_params)
        return ml_dataset.get_dataset(0)

    def write_data(
        self, data: xr.Dataset, data_id: str, replace: bool = False, **write_params
    ) -> str:
        assert_instance(data, xr.Dataset, name="data")
        return self.write_data_generic(data, data_id, replace=replace, **write_params)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import copy
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Tuple

import fsspec

from xcube.util.assertions import assert_instance
from xcube.util.jsonschema import (
    JsonBooleanSchema,
    JsonIntegerSchema,
    JsonNumberSchema,
    JsonObjectSchema,
)

from ..accessor import DataOpener, DataWriter
from ..datatype import DataType
from ..error import DataStoreError

COMMON_STORAGE_OPTIONS_SCHEMA_PROPERTIES = dict(
    # passed to ``DirCache``, if the implementation supports
    # directory listing caching. Pass use_listings_cache=False
    # to disable such caching.
    use_listings_cache=JsonBooleanSchema(),
    listings_expiry_time=JsonNumberSchema(),
    max_paths=JsonIntegerSchema(),
    # If this is a cachable implementation, pass True here to force
    # creating a new instance even if a matching instance exists, and prevent
    # storing this instance.
    skip_instance_cache=JsonBooleanSchema(),
    asynchronous=JsonBooleanSchema(),
)

PROTOCOL_PARAM_NAME = "protocol"
STORAGE_OPTIONS_PARAM_NAME = "storage_options"
FS_PARAM_NAME = "fs"
ROOT_PARAM_NAME = "root"


class FsAccessor:
    """
    Base class for accessing some filesystem.
    """

    @classmethod
    def get_protocol(cls) -> str:
        """Get the filesystem protocol."""
        return "abstract"

    @classmethod
    def get_storage_options_schema(cls) -> JsonObjectSchema:
        """Get the JSON schema of the filesystem parameters."""
        return JsonObjectSchema(
            properties=COMMON_STORAGE_OPTIONS_SCHEMA_PROPERTIES,
            additional_properties=True,
        )

    @classmethod
    def load_fs(
        cls, params: dict[str, Any]
    ) -> tuple[fsspec.AbstractFileSystem, Optional[str], dict[str, Any]]:
        """
        Load a filesystem instance from *params*.

        Pops useful parameters from a copy of *params* to get
        or instantiate the filesystem.
        Returns the filesystem and *params* reduced by
        the used parameters.

        Args:
            params: Parameters passed to a filesystem
                data store, opener, or writer call.

        Returns: A tuple comprising the filesystem, an optional root path,
            and the modified *params*.
        """
        params = dict(params)

        # Filesystem data-stores pass "fs" and "root" kwargs to
        # data opener and writer calls.
        fs = params.pop(FS_PARAM_NAME, None)
        root = params.pop(ROOT_PARAM_NAME, None)
        if fs is not None:
            assert_instance(fs, fsspec.AbstractFileSystem, name=FS_PARAM_NAME)
        if root is not None:
            assert_instance(root, str, name=ROOT_PARAM_NAME)
        if fs:
            return fs, root, params

        protocol = cls.get_protocol()
        if protocol == "abstract":
            protocol = params.pop(PROTOCOL_PARAM_NAME, None)
            if protocol is None:
                raise DataStoreError(
                    f"Cannot determine filesystem,"
                    f" try using parameter"
                    f" {PROTOCOL_PARAM_NAME!r}"
                )

        storage_options = params.pop(STORAGE_OPTIONS_PARAM_NAME, None)
        if storage_options is not None:
            assert_instance(storage_options, dict, name=STORAGE_OPTIONS_PARAM_NAME)
            storage_options = dict(storage_options)

        # Note, by default, filesystem data stores are writable and hence
        # SHALL NOT cache any directory listings!
        use_listings_cache = bool(
            storage_options.pop("use_listings_cache", False)
            if storage_options
            else False
        )

        try:
            return (
                fsspec.filesystem(
                    protocol,
                    use_listings_cache=use_listings_cache,
                    **(storage_options or {}),
                ),
                root,
                params,
            )
        except (ValueError, ImportError) as error:
            raise DataStoreError(f"Cannot instantiate filesystem {protocol!r}: {error}")

    @classmethod
    def add_storage_options_to_params_schema(
        cls, params_schema: JsonObjectSchema
    ) -> JsonObjectSchema:
        """
        Utility method to be used by subclasses to add the schema
        for the parameter "storage_options" to given *param_schema*.
        """
        params_schema = copy.deepcopy(params_schema)
        params_schema.properties[STORAGE_OPTIONS_PARAM_NAME] = (
            cls.get_storage_options_schema()
        )
        return params_schema

    @classmethod
    def remove_storage_options_from_params_schema(
        cls, params_schema: JsonObjectSchema
    ) -> JsonObjectSchema:
        """
        Utility method to be used by subclasses to remove the schema
        for the parameter "storage_options" from given *param_schema*.
        """
        if STORAGE_OPTIONS_PARAM_NAME in params_schema.properties:
            params_schema = copy.deepcopy(params_schema)
            del params_schema.properties[STORAGE_OPTIONS_PARAM_NAME]
        return params_schema


class FsDataAccessor(DataOpener, DataWriter, FsAccessor, ABC):
    """Abstract base class for data accessors that
    use an underlying filesystem.

    A ``FsDataAccessor`` is responsible for exactly one data type
    and one data format.

    Note, for all filesystem based accessors the parameter
    *data_id* is an absolute path that is build by the store
    according to ``f"{store.root}/{data_id}"``.
    The store's root may therefore only be useful to
    compute the original, relative store's *data_id*.
    """

    @classmethod
    @abstractmethod
    def get_data_type(cls) -> DataType:
        """Get the supported data type."""

    @classmethod
    @abstractmethod
    def get_format_id(cls) -> str:
        """Get the format identifier,
        for example "zarr" or "geojson".
        """

    def get_delete_data_params_schema(self, data_id: str = None) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                recursive=JsonBooleanSchema(),
                maxdepth=JsonIntegerSchema(),
                storage_options=self.get_storage_options_schema(),
            ),
            additional_properties=False,
        )

    def delete_data(self, data_id: str, **delete_params):
        fs, _, delete_params = self.load_fs(delete_params)
        fs.delete(data_id, **delete_params)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from collections.abc import Sequence
from typing import Any, Dict, Optional

from xcube.util.undefined import UNDEFINED


class GeoJSON:
    PRIMITIVE_GEOMETRY_TYPES = {
        "Point",
        "LineString",
        "Polygon",
        "MultiPoint",
        "MultiLineString",
        "MultiPolygon",
    }
    GEOMETRY_COLLECTION_TYPE = "GeometryCollection"
    FEATURE_TYPE = "Feature"
    FEATURE_COLLECTION_TYPE = "FeatureCollection"

    @classmethod
    def is_point(cls, obj: Any) -> bool:
        return cls.get_type_name(obj) == "Point"

    @classmethod
    def is_feature(cls, obj: Any) -> bool:
        return cls.get_type_name(obj) == cls.FEATURE_TYPE

    @classmethod
    def is_feature_collection(cls, obj: Any) -> bool:
        return cls.get_type_name(
            obj
        ) == cls.FEATURE_COLLECTION_TYPE and cls._is_valid_sequence(obj, "features")

    @classmethod
    def is_geometry(cls, obj: Any) -> bool:
        type_name = cls.get_type_name(obj)
        if type_name in cls.PRIMITIVE_GEOMETRY_TYPES:
            return cls._is_valid_sequence(obj, "coordinates")
        if type_name == cls.GEOMETRY_COLLECTION_TYPE:
            return cls._is_valid_sequence(obj, "geometries")
        return False

    @classmethod
    def is_geometry_collection(cls, obj: Any) -> bool:
        type_name = cls.get_type_name(obj)
        if type_name == cls.GEOMETRY_COLLECTION_TYPE:
            return cls._is_valid_sequence(obj, "geometries")
        return False

    @classmethod
    def get_geometry_collection_geometries(cls, obj: Any) -> Optional[Sequence[dict]]:
        type_name = cls.get_type_name(obj)
        if type_name == cls.GEOMETRY_COLLECTION_TYPE:
            geometries = cls._get_sequence(obj, "geometries")
            if geometries == UNDEFINED:
                return None
            elif geometries is None:
                return []
            else:
                return geometries
        return None

    @classmethod
    def get_feature_collection_features(cls, obj: Any) -> Optional[Sequence[dict]]:
        type_name = cls.get_type_name(obj)
        if type_name == cls.FEATURE_COLLECTION_TYPE:
            features = cls._get_sequence(obj, "features")
            if features == UNDEFINED:
                return None
            elif features is None:
                return []
            else:
                return features
        return None

    @classmethod
    def get_feature_geometry(cls, obj: Any) -> Optional[dict]:
        type_name = cls.get_type_name(obj)
        if type_name == cls.FEATURE_TYPE:
            if "geometry" not in obj:
                return None
            geometry = obj["geometry"]
            if cls.is_geometry(geometry):
                return geometry
            return None
        return None

    @classmethod
    def get_type_name(cls, obj: Any) -> Optional[str]:
        if not isinstance(obj, dict) or "type" not in obj:
            return None
        return obj["type"] or None

    @classmethod
    def _is_valid_sequence(cls, obj: Any, attr_name: str) -> bool:
        return cls._get_sequence(obj, attr_name) != UNDEFINED

    @classmethod
    def _get_sequence(cls, obj: Any, attr_name: str) -> Optional[Sequence[Any]]:
        if attr_name not in obj:
            return UNDEFINED
        sequence = obj[attr_name]
        if sequence is None:
            # GeoJSON sequence properties, 'coordinates', 'geometries', 'features', may be None according to spec
            return None
        try:
            iter(sequence)
            return sequence
        except TypeError:
            return UNDEFINED



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from .accessor import (
    DataDeleter,
    DataOpener,
    DataPreloader,
    DataTimeSliceUpdater,
    DataWriter,
    find_data_opener_extensions,
    find_data_writer_extensions,
    get_data_accessor_predicate,
    new_data_opener,
    new_data_writer,
)
from .assertions import assert_valid_config, assert_valid_params
from .datatype import (
    ANY_TYPE,
    DATASET_TYPE,
    GEO_DATA_FRAME_TYPE,
    MULTI_LEVEL_DATASET_TYPE,
    DataType,
    DataTypeLike,
)
from .descriptor import (
    DataDescriptor,
    DatasetDescriptor,
    GeoDataFrameDescriptor,
    MultiLevelDatasetDescriptor,
    VariableDescriptor,
    new_data_descriptor,
)
from .error import DataStoreError
from .fs.registry import new_fs_data_store
from .preload import PreloadHandle, PreloadState, PreloadStatus
from .search import DataSearcher, DefaultSearchMixin
from .store import (
    DataStore,
    MutableDataStore,
    PreloadedDataStore,
    find_data_store_extensions,
    get_data_store_class,
    get_data_store_params_schema,
    list_data_store_ids,
    new_data_store,
)
from .storepool import (
    DataStoreConfig,
    DataStoreInstance,
    DataStorePool,
    DataStorePoolLike,
    get_data_store_instance,
)
